@@ -5,9 +5,11 @@ Generic helpers for LLVM code generation.
 from __future__ import print_function, division, absolute_import
 from contextlib import contextmanager
 import functools
-from llvm.core import Constant, Type
-import llvm.core as lc
-from . import errcode
+import re
+from llvmlite.llvmpy.core import Constant, Type
+import llvmlite.llvmpy.core as lc
+
+from . import errcode, utils
 
 
 true_bit = Constant.int(Type.int(1), 1)
@@ -20,6 +22,10 @@ def as_bool_byte(builder, value):
     return builder.zext(value, Type.int(8))
 
 
+def as_bool_bit(builder, value):
+    return builder.icmp(lc.ICMP_NE, value, Constant.null(value.type))
+
+
 def make_anonymous_struct(builder, values):
     """
     Create an anonymous struct constant containing the given LLVM *values*.
@@ -37,7 +43,8 @@ class Structure(object):
     named fields and attribute access.
     """
 
-    def __init__(self, context, builder, value=None, ref=None):
+    # XXX Should this warrant several separate constructors?
+    def __init__(self, context, builder, value=None, ref=None, cast_ref=False):
         self._type = context.get_struct_type(self)
         self._context = context
         self._builder = builder
@@ -49,7 +56,14 @@ class Structure(object):
                 builder.store(value, self._value)
         else:
             assert value is None
-            assert self._type == ref.type.pointee
+            assert is_pointer(ref.type)
+            if self._type != ref.type.pointee:
+                if cast_ref:
+                    ref = builder.bitcast(ref, Type.pointer(self._type))
+                else:
+                    raise TypeError(
+                        "mismatching pointer type: got %s, expected %s"
+                        % (ref.type.pointee, self._type))
             self._value = ref
 
         self._namemap = {}
@@ -61,6 +75,13 @@ class Structure(object):
             self._fdmap.append((base, Constant.int(Type.int(), i)))
             self._typemap.append(tp)
 
+    def _get_ptr_by_index(self, index):
+        ptr = self._builder.gep(self._value, self._fdmap[index])
+        return ptr
+
+    def _get_ptr_by_name(self, attrname):
+        return self._get_ptr_by_index(self._namemap[attrname])
+
     def __getattr__(self, field):
         """
         Load the LLVM value of the named *field*.
@@ -82,18 +103,17 @@ class Structure(object):
         """
         Load the LLVM value of the field at *index*.
         """
-        offset = self._fdmap[index]
-        ptr = self._builder.gep(self._value, offset)
-        return self._builder.load(ptr)
+
+        return self._builder.load(self._get_ptr_by_index(index))
 
     def __setitem__(self, index, value):
         """
         Store the LLVM *value* into the field at *index*.
         """
-        offset = self._fdmap[index]
-        ptr = self._builder.gep(self._value, offset)
-        value = self._context.get_return_value(self._builder,
-                                               self._typemap[index], value)
+        ptr = self._get_ptr_by_index(index)
+        value = self._context.get_struct_member_value(self._builder,
+                                                      self._typemap[index],
+                                                      value)
         if ptr.type.pointee != value.type:
             raise AssertionError("Type mismatch: __setitem__(%d, ...) "
                                  "expected %r but got %r"
@@ -118,6 +138,12 @@ class Structure(object):
         """
         return self._builder.load(self._value)
 
+    def _setvalue(self, value):
+        """Store the value in this structure"""
+        assert not is_pointer(value.type)
+        assert value.type == self._type, (value.type, self._type)
+        self._builder.store(value, self._value)
+
     # __iter__ is derived by Python from __len__ and __getitem__
 
 
@@ -135,6 +161,10 @@ def append_basic_block(builder, name=''):
 
 @contextmanager
 def goto_block(builder, bb):
+    """
+    A context manager which temporarily positions *builder* at the end
+    of basic block *bb* (but before any terminator).
+    """
     bbold = builder.basic_block
     term = bb.terminator
     if term is not None:
@@ -152,9 +182,25 @@ def goto_entry_block(builder):
         yield
 
 
-def alloca_once(builder, ty, name=''):
+def alloca_once(builder, ty, size=None, name=''):
+    """Allocate stack memory at the entry block of the current function
+    pointed by ``builder`` withe llvm type ``ty``.  The optional ``size`` arg
+    set the number of element to allocate.  The default is 1.  The optional
+    ``name`` arg set the symbol name inside the llvm IR for debugging.
+    """
     with goto_entry_block(builder):
-        return builder.alloca(ty, name=name)
+        return builder.alloca(ty, size=size, name=name)
+
+
+def alloca_once_value(builder, value, name=''):
+    """
+    Like alloca_once(), but passing a *value* instead of a type.  The
+    type is inferred and the allocated slot is also initialized with the
+    given value.
+    """
+    storage = alloca_once(builder, value.type)
+    builder.store(value, storage)
+    return storage
 
 
 def terminate(builder, bbend):
@@ -177,10 +223,6 @@ def is_not_null(builder, val):
     return builder.icmp(lc.ICMP_NE, null, val)
 
 
-is_true = is_not_null
-is_false = is_null
-
-
 def set_branch_weight(builder, brinst, trueweight, falseweight):
     module = get_module(builder)
     mdid = lc.MetaDataString.get(module, "branch_weights")
@@ -211,8 +253,8 @@ def if_likely(builder, pred):
 @contextmanager
 def ifthen(builder, pred):
     bb = builder.basic_block
-    bbif = append_basic_block(builder, bb.name + '.if')
-    bbend = append_basic_block(builder, bb.name + '.endif')
+    bbif = append_basic_block(builder, add_postfix(bb.name, '.if'))
+    bbend = append_basic_block(builder, add_postfix(bb.name, '.endif'))
     builder.cbranch(pred, bbif, bbend)
 
     with goto_block(builder, bbif):
@@ -293,6 +335,56 @@ def for_range(builder, count, intp):
     builder.position_at_end(bbend)
 
 
+@contextmanager
+def for_range_slice(builder, start, stop, step, intp, inc=True):
+    """
+    Generate LLVM IR for a for-loop based on a slice
+
+    Parameters
+    -------------
+    builder : object
+        Builder object
+    start : int
+        The beginning value of the slice
+    stop : int
+        The end value of the slice
+    step : int
+        The step value of the slice
+    intp :
+        The data type
+    inc : boolean, optional
+        A flag to handle the step < 0 case, in which case we decrement the loop
+
+    Returns
+    -----------
+        None
+    """
+
+    bbcond = append_basic_block(builder, "for.cond")
+    bbbody = append_basic_block(builder, "for.body")
+    bbend = append_basic_block(builder, "for.end")
+    bbstart = builder.basic_block
+    builder.branch(bbcond)
+
+    with goto_block(builder, bbcond):
+        index = builder.phi(intp, name="loop.index")
+        if (inc):
+            pred = builder.icmp(lc.ICMP_SLT, index, stop)
+        else:
+            pred = builder.icmp(lc.ICMP_SGT, index, stop)
+        builder.cbranch(pred, bbbody, bbend)
+
+    with goto_block(builder, bbbody):
+        yield index
+        bbbody = builder.basic_block
+        incr = builder.add(index, step)
+        terminate(builder, bbcond)
+
+    index.add_incoming(start, bbstart)
+    index.add_incoming(incr, bbbody)
+    builder.position_at_end(bbend)
+
+
 @contextmanager
 def loop_nest(builder, shape, intp):
     with _loop_nest(builder, shape, intp) as indices:
@@ -380,10 +472,7 @@ def get_item_pointer2(builder, data, shape, strides, layout, inds,
         # Any layout
         dimoffs = [builder.mul(s, i) for s, i in zip(strides, indices)]
         offset = functools.reduce(builder.add, dimoffs)
-        base = builder.ptrtoint(data, offset.type)
-        where = builder.add(base, offset)
-        ptr = builder.inttoptr(where, data.type)
-        return ptr
+        return pointer_add(builder, data, offset)
 
 
 def normalize_slice(builder, slice, length):
@@ -408,29 +497,12 @@ def get_strides_from_slice(builder, ndim, strides, slice, ax):
     return builder.mul(slice.step, oldstrides[ax])
 
 
-class MetadataKeyStore(object):
-    def __init__(self, module, name):
-        self.module = module
-        self.key = name
-        self.nmd = self.module.get_or_insert_named_metadata("python.module")
-
-    def set(self, value):
-        """
-        Add a string value
-        """
-        md = lc.MetaData.get(self.module,
-                             [lc.MetaDataString.get(self.module, value)])
-        self.nmd.add(md)
-
-    def get(self):
-        """
-        Get string value
-        """
-        node = self.nmd._ptr.getOperand(0)
-        return lc._make_value(node.getOperand(0)).string
-
-
 def is_scalar_zero(builder, value):
+    """
+    Return a predicate representing whether *value* is equal to zero.
+    """
+    assert not is_pointer(value.type)
+    assert not is_struct(value.type)
     nullval = Constant.null(value.type)
     if value.type in (Type.float(), Type.double()):
         isnull = builder.fcmp(lc.FCMP_OEQ, nullval, value)
@@ -439,6 +511,48 @@ def is_scalar_zero(builder, value):
     return isnull
 
 
+def is_not_scalar_zero(builder, value):
+    """
+    Return a predicate representin whether a *value* is not equal to zero.
+    not exactly "not is_scalar_zero" because of nans
+    """
+    assert not is_pointer(value.type)
+    assert not is_struct(value.type)
+    nullval = Constant.null(value.type)
+    if value.type in (Type.float(), Type.double()):
+        isnull = builder.fcmp(lc.FCMP_UNE, nullval, value)
+    else:
+        isnull = builder.icmp(lc.ICMP_NE, nullval, value)
+    return isnull
+
+
+def is_scalar_zero_or_nan(builder, value):
+    """
+    Return a predicate representing whether *value* is equal to either zero
+    or NaN.
+    """
+    assert not is_pointer(value.type)
+    assert not is_struct(value.type)
+    nullval = Constant.null(value.type)
+    if value.type in (Type.float(), Type.double()):
+        isnull = builder.fcmp(lc.FCMP_UEQ, nullval, value)
+    else:
+        isnull = builder.icmp(lc.ICMP_EQ, nullval, value)
+    return isnull
+
+is_true = is_not_scalar_zero
+is_false = is_scalar_zero
+
+def is_scalar_neg(builder, value):
+    """is _value_ negative?. Assumes _value_ is signed"""
+    nullval = Constant.null(value.type)
+    if value.type in (Type.float(), Type.double()):
+        isneg = builder.fcmp(lc.FCMP_OLT, value, nullval)
+    else:
+        isneg = builder.icmp(lc.ICMP_SLT, value, nullval)
+    return isneg
+
+
 def guard_null(context, builder, value):
     with if_unlikely(builder, is_scalar_zero(builder, value)):
         context.return_errcode(builder, errcode.ASSERTION_ERROR)
@@ -448,14 +562,23 @@ guard_zero = guard_null
 
 
 def is_struct(ltyp):
+    """
+    Whether the LLVM type *typ* is a pointer type.
+    """
     return ltyp.kind == lc.TYPE_STRUCT
 
 
 def is_pointer(ltyp):
+    """
+    Whether the LLVM type *typ* is a struct type.
+    """
     return ltyp.kind == lc.TYPE_POINTER
 
 
 def is_struct_ptr(ltyp):
+    """
+    Whether the LLVM type *typ* is a pointer-to-struct type.
+    """
     return is_pointer(ltyp) and is_struct(ltyp.pointee)
 
 
@@ -510,6 +633,68 @@ def gep(builder, ptr, *inds):
     return builder.gep(ptr, idx)
 
 
+def pointer_add(builder, ptr, offset, return_type=None):
+    """
+    Add an integral *offset* to pointer *ptr*, and return a pointer
+    of *return_type* (or, if omitted, the same type as *ptr*).
+
+    Note the computation is done in bytes, and ignores the width of
+    the pointed item type.
+    """
+    intptr_t = Type.int(utils.MACHINE_BITS)
+    intptr = builder.ptrtoint(ptr, intptr_t)
+    if isinstance(offset, int):
+        offset = Constant.int(intptr_t, offset)
+    intptr = builder.add(intptr, offset)
+    return builder.inttoptr(intptr, return_type or ptr.type)
+
+
+def global_constant(builder_or_module, name, value, linkage=lc.LINKAGE_INTERNAL):
+    """
+    Get or create a (LLVM module-)global constant with *name* or *value*.
+    """
+    if isinstance(builder_or_module, lc.Module):
+        module = builder_or_module
+    else:
+        module = get_module(builder_or_module)
+    data = module.add_global_variable(value.type, name=name)
+    data.linkage = lc.LINKAGE_INTERNAL
+    data.global_constant = True
+    data.initializer = value
+    return data
+
+
+def divmod_by_constant(builder, val, divisor):
+    """
+    Compute the (quotient, remainder) of *val* divided by the constant
+    positive *divisor*.  The semantics reflects those of Python integer
+    floor division, rather than C's / LLVM's signed division and modulo.
+    The difference lies with a negative *val*.
+    """
+    assert divisor > 0
+    divisor = Constant.int(val.type, divisor)
+    one = Constant.int(val.type, 1)
+
+    quot = alloca_once(builder, val.type)
+
+    with ifelse(builder, is_neg_int(builder, val)) as (if_neg, if_pos):
+        with if_pos:
+            # quot = val / divisor
+            quot_val = builder.sdiv(val, divisor)
+            builder.store(quot_val, quot)
+        with if_neg:
+            # quot = -1 + (val + 1) / divisor
+            val_plus_one = builder.add(val, one)
+            quot_val = builder.sdiv(val_plus_one, divisor)
+            builder.store(builder.sub(quot_val, one), quot)
+
+    # rem = val - quot * divisor
+    # (should be slightly faster than a separate modulo operation)
+    quot_val = builder.load(quot)
+    rem_val = builder.sub(val, builder.mul(quot_val, divisor))
+    return quot_val, rem_val
+
+
 # ------------------------------------------------------------------------------
 # Debug
 
@@ -553,8 +738,39 @@ def printf(builder, format_string, *values):
             args.append(Constant.int(Type.int(), v))
         elif isinstance(v, float):
             args.append(Constant.real(Type.double(), v))
-
+        else:
+            args.append(v)
     functype = Type.function(Type.int(32), [Type.pointer(Type.int(8))], True)
-    fn = get_module(builder).add_function(functype, 'printf')
+    fn = get_module(builder).get_or_insert_function(functype, 'printf')
     builder.call(fn, [str_addr] + args)
 
+
+def cbranch_or_continue(builder, cond, bbtrue):
+    """
+    Branch conditionally or continue.
+
+    Note: a new block is created and builder is moved to the end of the new
+          block.
+    """
+    fn = get_function(builder)
+    bbcont = fn.append_basic_block('.continue')
+    builder.cbranch(cond, bbtrue, bbcont)
+    builder.position_at_end(bbcont)
+    return bbcont
+
+
+def add_postfix(name, postfix):
+    """Add postfix to string.  If the postfix is already there, add a counter.
+    """
+    regex = "(.*{0})([0-9]*)$".format(postfix)
+    m = re.match(regex, name)
+    if m:
+        head, ct = m.group(1), m.group(2)
+        if len(ct):
+            ct = int(ct) + 1
+        else:
+            ct = 1
+
+        return "{head}{ct}".format(head=head, ct=ct)
+    return name + postfix
+
