@@ -13,6 +13,7 @@ IS_32BITS = MACHINE_BITS == 32
 class NumbaWarning(Warning):
     pass
 
+
 def _readenv(name, ctor, default):
     try:
         res = os.environ[name]
@@ -80,6 +81,11 @@ PYVERSION = sys.version_info[:2]
 # Disable CUDA support
 DISABLE_CUDA = _readenv("NUMBA_DISABLE_CUDA", int, 0)
 
+# Allow interpreter fallback so that Numba @jit decorator will never fail
+# Use for migrating from old numba (<0.12) which supported closure, and other
+# yet-to-be-supported features.
+COMPATIBILITY_MODE = _readenv("NUMBA_COMPATIBILITY_MODE", int, 0)
+
 # Force CUDA compute capability
 def _force_cc(text):
     if not text:
@@ -95,3 +101,7 @@ def _force_cc(text):
 
 
 FORCE_CUDA_CC = _readenv("NUMBA_FORCE_CUDA_CC", _force_cc, None)
+
+# x86-64 specific
+# Enable AVX on supported platforms.
+ENABLE_AVX = _readenv("NUMBA_ENABLE_AVX", int, 0)
