@@ -6,6 +6,8 @@ strides, dtype and size attributes similar to a NumPy ndarray.
 from __future__ import print_function, absolute_import, division
 import warnings
 import math
+import copy
+from ctypes import c_void_p
 import numpy as np
 from .ndarray import (ndarray_populate_head, ArrayHeaderManager)
 from . import driver as _driver
@@ -86,24 +88,37 @@ class DeviceNDArrayBase(object):
         self.dtype = np.dtype(dtype)
         self.size = int(np.prod(self.shape))
         # prepare gpu memory
-        if gpu_data is None:
-            self.alloc_size = _driver.memory_size_from_info(self.shape,
-                                                            self.strides,
-                                                            self.dtype.itemsize)
-            gpu_data = devices.get_context().memalloc(self.alloc_size)
+        if self.size > 0:
+            if gpu_data is None:
+                self.alloc_size = _driver.memory_size_from_info(self.shape,
+                                                                self.strides,
+                                                                self.dtype.itemsize)
+                gpu_data = devices.get_context().memalloc(self.alloc_size)
+            else:
+                self.alloc_size = _driver.device_memory_size(gpu_data)
         else:
-            self.alloc_size = _driver.device_memory_size(gpu_data)
+            gpu_data = None
+            self.alloc_size = 0
 
         self.gpu_mem = ArrayHeaderManager(devices.get_context())
 
         if gpu_head is None:
-            gpu_head = self.gpu_mem.allocate(self.ndim)
-            ndarray_populate_head(gpu_head, gpu_data, self.shape,
-                                  self.strides, stream=stream)
+            gpu_head = ndarray_populate_head(self.gpu_mem, gpu_data,
+                                             self.shape, self.strides,
+                                             stream=stream)
         self.gpu_head = gpu_head
         self.gpu_data = gpu_data
 
-        self.__writeback = writeback # should deprecate the use of this
+        self.__writeback = writeback    # should deprecate the use of this
+        self.stream = 0
+
+    def bind(self, stream=0):
+        """Bind a CUDA stream to this object so that all subsequent operation
+        on this array defaults to the given stream.
+        """
+        clone = copy.copy(self)
+        clone.stream = stream
+        return clone
 
     def __del__(self):
         try:
@@ -111,6 +126,9 @@ class DeviceNDArrayBase(object):
         except:
             pass
 
+    def _default_stream(self, stream):
+        return self.stream if not stream else stream
+
     @property
     def _numba_type_(self):
         """
@@ -124,7 +142,11 @@ class DeviceNDArrayBase(object):
     def device_ctypes_pointer(self):
         """Returns the ctypes pointer to the GPU data buffer
         """
-        return self.gpu_data.device_ctypes_pointer
+        if self.gpu_data is None:
+            return c_void_p(0)
+        else:
+            return self.gpu_data.device_ctypes_pointer
+
 
     def copy_to_device(self, ary, stream=0):
         """Copy `ary` to `self`.
@@ -132,9 +154,12 @@ class DeviceNDArrayBase(object):
         If `ary` is a CUDA memory, perform a device-to-device transfer.
         Otherwise, perform a a host-to-device transfer.
         """
+        if ary.size == 0:
+            # Nothing to do
+            return
+        stream = self._default_stream(stream)
         if _driver.is_device_memory(ary):
-            sz = min(_driver.device_memory_size(self),
-                     _driver.device_memory_size(ary))
+            sz = min(self.alloc_size, ary.alloc_size)
             _driver.device_to_device(self, ary, sz, stream=stream)
         else:
             sz = min(_driver.host_memory_size(ary), self.alloc_size)
@@ -146,6 +171,7 @@ class DeviceNDArrayBase(object):
 
         Always returns the host array.
         """
+        stream = self._default_stream(stream)
         if ary is None:
             hostary = np.empty(shape=self.alloc_size, dtype=np.byte)
         else:
@@ -164,7 +190,10 @@ class DeviceNDArrayBase(object):
                     raise TypeError('incompatible strides; device %s; host %s' %
                                     (self.strides, ary.strides))
             hostary = ary
-        _driver.device_to_host(hostary, self, self.alloc_size, stream=stream)
+
+        assert self.alloc_size >= 0, "Negative memory size"
+        if self.alloc_size != 0:
+            _driver.device_to_host(hostary, self, self.alloc_size, stream=stream)
 
         if ary is None:
             hostary = np.ndarray(shape=self.shape, strides=self.strides,
@@ -172,6 +201,7 @@ class DeviceNDArrayBase(object):
         return hostary
 
     def to_host(self, stream=0):
+        stream = self._default_stream(stream)
         warnings.warn("to_host() is deprecated and will be removed",
                       DeprecationWarning)
         if self.__writeback is None:
@@ -183,6 +213,7 @@ class DeviceNDArrayBase(object):
         If the array cannot be equally divided, the last section will be
         smaller.
         """
+        stream = self._default_stream(stream)
         if self.ndim != 1:
             raise ValueError("only support 1d array")
         if self.strides[0] != self.dtype.itemsize:
@@ -216,7 +247,7 @@ class DeviceNDArray(DeviceNDArrayBase):
 
         Reshape the array and keeping the original data
         """
-        if len(newshape) == 1 and isinstance(newshape, (tuple, list)):
+        if len(newshape) == 1 and isinstance(newshape[0], (tuple, list)):
             newshape = newshape[0]
 
         cls = type(self)
@@ -234,6 +265,7 @@ class DeviceNDArray(DeviceNDArrayBase):
             raise NotImplementedError("operation requires copying")
 
     def ravel(self, order='C', stream=0):
+        stream = self._default_stream(stream)
         cls = type(self)
         newarr, extents = self._dummy.ravel(order=order)
 
@@ -246,24 +278,36 @@ class DeviceNDArray(DeviceNDArrayBase):
             raise NotImplementedError("operation requires copying")
 
     def __getitem__(self, item):
+        return self._do_getitem(item)
+
+    def getitem(self, item, stream=0):
+        """Do `__getitem__(item)` with CUDA stream
+        """
+        return self._do_getitem(item, stream)
+
+    def _do_getitem(self, item, stream=0):
+        stream = self._default_stream(stream)
+
         arr = self._dummy.__getitem__(item)
         extents = list(arr.iter_contiguous_extent())
         cls = type(self)
         if len(extents) == 1:
             newdata = self.gpu_data.view(*extents[0])
 
-            if dummyarray.is_element_indexing(item, self.ndim):
+            if not arr.is_array:
+                # Element indexing
                 hostary = np.empty(1, dtype=self.dtype)
                 _driver.device_to_host(dst=hostary, src=newdata,
-                                       size=self._dummy.itemsize)
+                                       size=self._dummy.itemsize,
+                                       stream=stream)
                 return hostary[0]
             else:
                 return cls(shape=arr.shape, strides=arr.strides,
-                           dtype=self.dtype, gpu_data=newdata)
+                           dtype=self.dtype, gpu_data=newdata, stream=stream)
         else:
             newdata = self.gpu_data.view(*arr.extent)
             return cls(shape=arr.shape, strides=arr.strides,
-                       dtype=self.dtype, gpu_data=newdata)
+                       dtype=self.dtype, gpu_data=newdata, stream=stream)
 
 
 class MappedNDArray(DeviceNDArrayBase, np.ndarray):
@@ -273,11 +317,8 @@ class MappedNDArray(DeviceNDArrayBase, np.ndarray):
 
     def device_setup(self, gpu_data, stream=0):
         self.gpu_mem = ArrayHeaderManager(devices.get_context())
-
-        gpu_head = self.gpu_mem.allocate(self.ndim)
-        ndarray_populate_head(gpu_head, gpu_data, self.shape,
-                              self.strides, stream=stream)
-
+        gpu_head = ndarray_populate_head(self.gpu_mem, gpu_data, self.shape,
+                                         self.strides, stream=stream)
         self.gpu_data = gpu_data
         self.gpu_head = gpu_head
 
