@@ -4,6 +4,16 @@ import numba.ctypes_support as ctypes
 import contextlib
 from collections import deque
 from . import devices, driver
+from numba.targets.arrayobj import make_array_ctype
+from numba.targets.registry import CPUTarget
+from numba import types
+
+
+def _calc_array_sizeof(ndim):
+    """Use the ABI size in the CPU target
+    """
+    ctx = CPUTarget.target_context
+    return ctx.calc_array_sizeof(ndim)
 
 
 class ArrayHeaderManager(object):
@@ -30,7 +40,7 @@ class ArrayHeaderManager(object):
 
     # Maximum size for each array head
     #    = 4 (ndim) * 8 (sizeof intp) * 2 (shape strides) + 8 (ptr)
-    elemsize = 4 * 8 * 2 + 8
+    elemsize = _calc_array_sizeof(4)
 
     # Number of page-locked staging area
     num_stages = 5
@@ -107,19 +117,6 @@ class ArrayHeaderManager(object):
         return "<cuda managed memory %s >" % (self.context.device,)
 
 
-def make_array_ctype(ndim):
-    """Create a array header type for a given dimension.
-    """
-    c_intp = ctypes.c_ssize_t
-
-    class c_array(ctypes.Structure):
-        _fields_ = [('data', ctypes.c_void_p),
-                    ('shape', c_intp * ndim),
-                    ('strides', c_intp * ndim)]
-
-    return c_array
-
-
 def _allocate_head(nd):
     """Allocate the metadata structure
     """
@@ -146,7 +143,8 @@ def ndarray_populate_head(gpu_mem, gpu_data, shape, strides, stream=0):
     assert nd > 0, "0 or negative dimension"
 
     arraytype = make_array_ctype(nd)
-    struct = arraytype(data=driver.device_pointer(gpu_data),
+    struct = arraytype(parent=None,
+                       data=driver.device_pointer(gpu_data),
                        shape=shape,
                        strides=strides)
 
