@@ -5,6 +5,7 @@ from __future__ import print_function, absolute_import, division
 import sys, logging, re
 from ctypes import (c_void_p, c_int, POINTER, c_char_p, c_size_t, byref,
                     c_char)
+from numba import config
 from .error import NvvmError, NvvmSupportError
 from .libs import open_libdevice, open_cudalib
 
@@ -89,7 +90,7 @@ class NVVM(object):
             except OSError as e:
                 cls.__INSTANCE = None
                 errmsg = ("libNVVM cannot be found. Do `conda install "
-                          "cudatoolkit`")
+                          "cudatoolkit`:\n%s")
                 raise NvvmSupportError(errmsg % e)
 
             # Find & populate functions
@@ -211,6 +212,9 @@ class CompilationUnit(object):
         # get log
         self.log = self.get_log()
 
+        if config.DUMP_ASSEMBLY:
+            print(ptxbuf[:].decode('ascii'))
+
         return ptxbuf[:]
 
     def _try_error(self, err, msg):
@@ -239,22 +243,38 @@ data_layout = {
 
 default_data_layout = data_layout[tuple.__itemsize__ * 8]
 
-SUPPORTED_CC = frozenset([(2, 0), (3, 0), (3, 5)])
+
+# List of supported compute capability in sorted order
+SUPPORTED_CC = (2, 0), (3, 0), (3, 5)
+
+
+def _find_arch(mycc):
+    for i, cc in enumerate(SUPPORTED_CC):
+        if cc == mycc:
+            # Matches
+            return cc
+        elif cc > mycc:
+            # Exceeded
+            if i == 0:
+                # CC lower than supported
+                raise NvvmSupportError("GPU compute capability %d.%d is "
+                                       "not supported (requires >=2.0)" % mycc)
+            else:
+                # return the previous CC
+                return SUPPORTED_CC[i - 1]
+
+    # CC higher than supported
+    return SUPPORTED_CC[-1]   # Choose the highest
 
 
 def get_arch_option(major, minor):
-    if major == 2:
-        minor = 0
-    if major == 3:
-        if minor < 5:
-            minor = 0
-        else:
-            minor = 5
-    if (major, minor) not in SUPPORTED_CC:
-        raise Exception("compute compability %d.%d is not supported" %
-                        (major, minor))
-    arch = 'compute_%d%d' % (major, minor)
-    return arch
+    """Matches with the closest architecture option
+    """
+    if config.FORCE_CUDA_CC:
+        arch = config.FORCE_CUDA_CC
+    else:
+        arch = _find_arch((major, minor))
+    return 'compute_%d%d' % arch
 
 
 MISSING_LIBDEVICE_MSG = '''
@@ -279,9 +299,25 @@ class LibDevice(object):
         return self.bc
 
 
+ir_numba_cas_hack = """
+define internal i32 @___numba_cas_hack(i32* %ptr, i32 %cmp, i32 %val) alwaysinline {
+    %out = cmpxchg volatile i32* %ptr, i32 %cmp, i32 %val monotonic
+    ret i32 %out
+}
+"""
+
+
 def llvm_to_ptx(llvmir, **opts):
     cu = CompilationUnit()
     libdevice = LibDevice(arch=opts.get('arch', 'compute_20'))
+    # New LLVM generate a shorthand for datalayout that NVVM does not know
+    llvmir = llvmir.replace('e-i64:64-v16:16-v32:32-n16:32:64',
+                            default_data_layout)
+    # Replace with our cmpxchg implementation because LLVM 3.5 has a new
+    # semantic for cmpxchg.
+    llvmir = llvmir.replace('declare i32 @___numba_cas_hack(i32*, i32, i32)',
+                            ir_numba_cas_hack)
+
     llvmir = llvm33_to_32_ir(llvmir)
     cu.add_module(llvmir.encode('utf8'))
     cu.add_module(libdevice.get())
@@ -296,10 +332,13 @@ re_fnattr_def = re.compile('attributes\s+(#\d+)\s*=\s*{((?:\s*\w+)+)\s*}')
 def llvm33_to_32_ir(ir):
     """rewrite function attributes in the IR
     """
+
+    invalid_attrs = frozenset(['noduplicate'])
+
     attrs = {}
     for m in re_fnattr_def.finditer(ir):
         ct, text = m.groups()
-        attrs[ct] = text
+        attrs[ct] = ' '.join(set(text.split()) - invalid_attrs)
 
     def scanline(line):
         if line.startswith('define') or line.startswith('declare'):
@@ -314,7 +353,7 @@ def llvm33_to_32_ir(ir):
 
 
 def set_cuda_kernel(lfunc):
-    from llvm.core import MetaData, MetaDataString, Constant, Type
+    from llvmlite.llvmpy.core import MetaData, MetaDataString, Constant, Type
 
     m = lfunc.module
 
