@@ -4,8 +4,8 @@ This scripts specifies all PTX special objects.
 from __future__ import print_function, absolute_import, division
 import operator
 import numpy
-import llvm.core as lc
-from numba import types, ir, typing, macro, dispatcher
+import llvmlite.llvmpy.core as lc
+from numba import types, ir, typing, macro
 from .cudadrv import nvvm
 
 
@@ -143,20 +143,21 @@ class syncthreads(Stub):
 #-------------------------------------------------------------------------------
 # shared
 
+def _legalize_shape(shape):
+    if isinstance(shape, tuple):
+        return shape
+    elif isinstance(shape, int):
+        return (shape,)
+    else:
+        raise TypeError("invalid type for shape; got {0}".format(type(shape)))
+
 
 def shared_array(shape, dtype):
-    ndim = 1
-    if isinstance(shape, tuple):
-        ndim = len(shape)
-
+    shape = _legalize_shape(shape)
+    ndim = len(shape)
     fname = "ptx.smem.alloc"
     restype = types.Array(dtype, ndim, 'C')
-    if ndim == 1:
-        sig = typing.signature(restype, types.intp, types.Any)
-    else:
-        sig = typing.signature(restype, types.UniTuple(types.intp, ndim),
-                               types.Any)
-
+    sig = typing.signature(restype, types.UniTuple(types.intp, ndim), types.Any)
     return ir.Intrinsic(fname, sig, args=(shape, dtype))
 
 
@@ -174,18 +175,11 @@ class shared(Stub):
 
 
 def local_array(shape, dtype):
-    ndim = 1
-    if isinstance(shape, tuple):
-        ndim = len(shape)
-
+    shape = _legalize_shape(shape)
+    ndim = len(shape)
     fname = "ptx.lmem.alloc"
     restype = types.Array(dtype, ndim, 'C')
-    if ndim == 1:
-        sig = typing.signature(restype, types.intp, types.Any)
-    else:
-        sig = typing.signature(restype, types.UniTuple(types.intp, ndim),
-                               types.Any)
-
+    sig = typing.signature(restype, types.UniTuple(types.intp, ndim), types.Any)
     return ir.Intrinsic(fname, sig, args=(shape, dtype))
 
 
@@ -203,7 +197,10 @@ class local(Stub):
 
 def const_array_like(ndarray):
     fname = "ptx.cmem.arylike"
-    aryty = dispatcher.Overloaded.typeof_pyval(ndarray)
+
+    from .descriptor import CUDATargetDesc
+    aryty = CUDATargetDesc.typingctx.resolve_argument_type(ndarray)
+
     sig = typing.signature(aryty, aryty)
     return ir.Intrinsic(fname, sig, args=[ndarray])
 
