@@ -1,12 +1,14 @@
 from __future__ import print_function, absolute_import
 import re
-from llvm.core import (Type, Builder, LINKAGE_INTERNAL, inline_function,
+from llvmlite.llvmpy.core import (Type, Builder, LINKAGE_INTERNAL,
                        Constant, ICMP_EQ)
-import llvm.passes as lp
+import llvmlite.llvmpy.core as lc
+import llvmlite.binding as ll
+
 from numba import typing, types, cgutils
 from numba.targets.base import BaseContext
 from .cudadrv import nvvm
-from . import nvvmutils
+from . import codegen, nvvmutils
 
 
 # -----------------------------------------------------------------------------
@@ -28,12 +30,23 @@ VALID_CHARS = re.compile(r'[^a-z0-9]', re.I)
 
 class CUDATargetContext(BaseContext):
     implement_powi_as_math_call = True
+    strict_alignment = True
 
     def init(self):
         from . import cudaimpl, libdevice
 
+        self._internal_codegen = codegen.JITCUDACodegen("numba.cuda.jit")
+
         self.insert_func_defn(cudaimpl.registry.functions)
         self.insert_func_defn(libdevice.registry.functions)
+        self._target_data = ll.create_target_data(nvvm.default_data_layout)
+
+    def jit_codegen(self):
+        return self._internal_codegen
+
+    @property
+    def target_data(self):
+        return self._target_data
 
     def mangler(self, name, argtypes):
         def repl(m):
@@ -47,27 +60,26 @@ class CUDATargetContext(BaseContext):
     def prepare_cuda_kernel(self, func, argtypes):
         # Adapt to CUDA LLVM
         module = func.module
-        func.linkage = LINKAGE_INTERNAL
         wrapper = self.generate_kernel_wrapper(func, argtypes)
-        func.delete()
-        del func
-
-        nvvm.set_cuda_kernel(wrapper)
+        func.linkage = LINKAGE_INTERNAL
         nvvm.fix_data_layout(module)
-
         return wrapper
 
     def generate_kernel_wrapper(self, func, argtypes):
         module = func.module
-        argtys = self.get_arguments(func.type.pointee)
-        fnty = Type.function(Type.void(), argtys)
-        wrapfn = module.add_function(fnty, name="cudaPy_" + func.name)
+        argtys = [self.get_argument_type(ty) for ty in argtypes]
+        wrapfnty = Type.function(Type.void(), argtys)
+        wrapper_module = self.create_module("cuda.kernel.wrapper")
+        fnty = Type.function(Type.int(),
+                             [self.get_return_type(types.pyobject)] + argtys)
+        func = wrapper_module.add_function(fnty, name=func.name)
+        wrapfn = wrapper_module.add_function(wrapfnty, name="cudaPy_" + func.name)
         builder = Builder.new(wrapfn.append_basic_block(''))
 
         # Define error handling variables
         def define_error_gv(postfix):
-            gv = module.add_global_variable(Type.int(),
-                                            name=wrapfn.name + postfix)
+            gv = wrapper_module.add_global_variable(Type.int(),
+                                                    name=wrapfn.name + postfix)
             gv.initializer = Constant.null(gv.type.pointee)
             return gv
 
@@ -96,8 +108,13 @@ class CUDATargetContext(BaseContext):
 
             # Use atomic cmpxchg to prevent rewriting the error status
             # Only the first error is recorded
-            xchg = builder.atomic_cmpxchg(gv_exc, old, status.code,
-                                          "monotonic")
+
+            casfnty = lc.Type.function(old.type, [gv_exc.type, old.type,
+                                                  old.type])
+
+            casfn = wrapper_module.add_function(casfnty,
+                                                name="___numba_cas_hack")
+            xchg = builder.call(casfn, [gv_exc, old, status.code])
             changed = builder.icmp(ICMP_EQ, xchg, old)
 
             # If the xchange is successful, save the thread ID.
@@ -112,16 +129,14 @@ class CUDATargetContext(BaseContext):
                     builder.store(val, ptr)
 
         builder.ret_void()
-
         # force inline
-        inline_function(status.code)
-
+        # inline_function(status.code)
+        nvvm.set_cuda_kernel(wrapfn)
+        module.link_in(ll.parse_assembly(str(wrapper_module)))
         module.verify()
-        return wrapfn
 
-    def link_dependencies(self, module, depends):
-        for lib in depends:
-            module.link_in(lib, preserve=True)
+        wrapfn = module.get_function(wrapfn.name)
+        return wrapfn
 
     def make_constant_array(self, builder, typ, ary):
         """
@@ -149,14 +164,14 @@ class CUDATargetContext(BaseContext):
             if gv.name == name and gv.type.pointee == text.type:
                 break
         else:
-            gl = lmod.add_global_variable(text.type, name=name,
+            gv = lmod.add_global_variable(text.type, name=name,
                                           addrspace=nvvm.ADDRSPACE_CONSTANT)
-            gl.linkage = LINKAGE_INTERNAL
-            gl.global_constant = True
-            gl.initializer = text
+            gv.linkage = LINKAGE_INTERNAL
+            gv.global_constant = True
+            gv.initializer = text
 
-            constcharptrty = Type.pointer(charty, nvvm.ADDRSPACE_CONSTANT)
-            charptr = builder.bitcast(gl, constcharptrty)
+        constcharptrty = Type.pointer(charty, nvvm.ADDRSPACE_CONSTANT)
+        charptr = builder.bitcast(gv, constcharptrty)
 
         conv = nvvmutils.insert_addrspace_conv(lmod, charty,
                                                nvvm.ADDRSPACE_CONSTANT)
@@ -165,10 +180,12 @@ class CUDATargetContext(BaseContext):
     def optimize_function(self, func):
         """Run O1 function passes
         """
-        fpm = lp.FunctionPassManager.new(func.module)
-
-        lp.PassManagerBuilder.new().populate(fpm)
-
-        fpm.initialize()
-        fpm.run(func)
-        fpm.finalize()
+        pass
+        ## XXX skipped for now
+        # fpm = lp.FunctionPassManager.new(func.module)
+        #
+        # lp.PassManagerBuilder.new().populate(fpm)
+        #
+        # fpm.initialize()
+        # fpm.run(func)
+        # fpm.finalize()
