@@ -1,14 +1,11 @@
 from __future__ import print_function, division, absolute_import
 
-try:
-    import __builtin__ as builtins
-except ImportError:
-    import builtins
 import collections
 import dis
 import sys
 
 from numba import ir, controlflow, dataflow, utils
+from numba.utils import builtins
 
 
 class Assigner(object):
@@ -82,15 +79,14 @@ class Interpreter(object):
 
         # { inst offset : ir.Block }
         self.blocks = {}
-        self.syntax_info = []
+        # { name: value } of global variables used by the bytecode
+        self.used_globals = {}
 
         # Temp states during interpretation
-
         self.current_block = None
         self.current_block_offset = None
         self.syntax_blocks = []
         self.dfainfo = None
-        self._block_actions = {}
         self.constants = {}
 
     def _fill_global_scope(self, scope):
@@ -98,6 +94,12 @@ class Interpreter(object):
         """
         pass
 
+    def get_used_globals(self):
+        """
+        Return a dictionary of global variables used by the bytecode.
+        """
+        return self.used_globals
+
     def _fill_args_into_scope(self, scope):
         for arg in self.argspec.args:
             scope.define(name=arg, loc=self.loc)
@@ -111,15 +113,111 @@ class Interpreter(object):
         # Interpret loop
         for inst, kws in self._iter_inst():
             self._dispatch(inst, kws)
-            # Clean up
-        self._remove_invalid_syntax_blocks()
+        # Clean up
+        self._insert_var_dels()
 
-    def _remove_invalid_syntax_blocks(self):
-        self.syntax_info = [syn for syn in self.syntax_info if syn.valid()]
+    def _insert_var_dels(self):
+        """
+        Insert ir.Del statements where necessary for the various
+        variables defined within the IR.
+        """
+        # { var name -> { block offset -> stmt of last use } }
+        var_use = collections.defaultdict(dict)
+        cfg = self.cfa.graph
 
-    def verify(self):
-        for b in utils.dict_itervalues(self.blocks):
-            b.verify()
+        for offset, ir_block in self.blocks.items():
+            for stmt in ir_block.body:
+                for var_name in stmt.list_vars():
+                    if stmt.is_terminator and not isinstance(stmt, ir.Return):
+                        # Move the "last use" at the beginning of successor
+                        # blocks.
+                        for succ, _ in cfg.successors(offset):
+                            var_use[var_name.name].setdefault(succ, None)
+                    else:
+                        var_use[var_name.name][offset] = stmt
+
+        for name in sorted(var_use):
+            var_use_map = var_use[name]
+            blocks = self._compute_var_disposal(name, var_use_map)
+            self._patch_var_disposal(name, var_use_map, blocks)
+
+    def _patch_var_disposal(self, var_name, use_map, disposal_blocks):
+        """
+        Insert ir.Del statements into each of the *disposal_blocks*
+        for variable *var_name*, at the appropriate points in the blocks.
+        """
+        for b in disposal_blocks:
+            # Either this is an original use point and we will insert after
+            # that statement, or it's a computed one and we will insert
+            # at the beginning of the block.
+            stmt = use_map.get(b)
+            ir_block = self.blocks[b]
+            if stmt is None:
+                ir_block.prepend(ir.Del(var_name, loc=ir_block.loc))
+            else:
+                # If the variable is used in a Return statement, we
+                # don't insert anything.
+                if not isinstance(stmt, ir.Return):
+                    ir_block.insert_after(ir.Del(var_name, loc=stmt.loc), stmt)
+
+    def _compute_var_disposal(self, var_name, use_map):
+        """
+        Compute a set of blocks in which *var_name* should be disposed of.
+        """
+        cfg = self.cfa.graph
+
+        # Compute a set of blocks where we want the variable to be
+        # kept alive.
+        live_points = set()
+
+        # Find the loops the variable is used in
+        loops = set()
+        enclosing_loops = None
+        for block in use_map:
+            block_loops = cfg.in_loops(block)
+            loops.update(block_loops)
+            if enclosing_loops is None:
+                enclosing_loops = set(block_loops)
+            else:
+                enclosing_loops.intersection_update(block_loops)
+
+        spanned_loops = loops - enclosing_loops
+
+        for loop in spanned_loops:
+            # As there are blocks outside the loop, the variable must be
+            # kept alive accross loop iterations => we only consider
+            # loop exit points.
+            # (otherwise it means the variable is *defined* inside the loop
+            #  and need not be kept alive accross iterations)
+            live_points.update(loop.exits)
+
+        for block in use_map:
+            live_points.add(block)
+
+        # Now compute a set of *exclusive* blocks covering all paths
+        # stemming from the live points. This ensures that each variable
+        # is del'ed at most once per possible code path.
+        tails = set()
+        # We use the todo list as a stack, so we really iterate in
+        # reverse topo order.
+        todo = list(cfg.topo_sort(live_points))
+        seen = set()
+        while todo:
+            b = todo.pop()
+            if b in seen:
+                continue
+            seen.add(b)
+            block_loops = cfg.in_loops(b)
+            if block_loops and block_loops[0] not in enclosing_loops:
+                continue
+            if cfg.descendents(b) & tails:
+                for succ, _ in cfg.successors(b):
+                    if succ not in tails:
+                        todo.append(succ)
+            else:
+                tails.add(b)
+
+        return tails
 
     def init_first_block(self):
         # Duplicate arguments so that these values can be casted into different
@@ -150,12 +248,9 @@ class Interpreter(object):
         if oldblock is not None and not oldblock.is_terminated:
             jmp = ir.Jump(inst.offset, loc=self.loc)
             oldblock.append(jmp)
-            # Get DFA block info
+        # Get DFA block info
         self.dfainfo = self.dfa.infos[self.current_block_offset]
         self.assigner = Assigner()
-        # Notify listeners for the new block
-        for fn in utils.dict_itervalues(self._block_actions):
-            fn(self.current_block_offset, self.current_block)
 
     def _end_current_block(self):
         self._remove_unused_temporaries()
@@ -196,7 +291,7 @@ class Interpreter(object):
         as a builtins (second).  If both failed, return a ir.UNDEFINED.
         """
         try:
-            return utils.func_globals(self.bytecode.func)[name]
+            return utils.get_function_globals(self.bytecode.func)[name]
         except KeyError:
             return getattr(builtins, name, ir.UNDEFINED)
 
@@ -234,12 +329,17 @@ class Interpreter(object):
         except AttributeError:
             raise NotImplementedError(inst)
         else:
-            return fn(inst, **kws)
+            try:
+                return fn(inst, **kws)
+            except ir.NotDefinedError as e:
+                if e.loc is None:
+                    e.loc = self.loc
+                raise e
 
     def dump(self, file=None):
         file = file or sys.stdout
         for offset, block in sorted(self.blocks.items()):
-            print('label %d:' % offset, file=file)
+            print('label %s:' % (offset,), file=file)
             block.dump(file=file)
 
     # --- Scope operations ---
@@ -259,6 +359,9 @@ class Interpreter(object):
         self.current_block.append(stmt)
 
     def get(self, name):
+        """
+        Get the variable (a Var instance) with the given *name*.
+        """
         # Try to simplify the variable lookup by returning an earlier
         # variable assigned to *name*.
         var = self.assigner.get_assignment_source(name)
@@ -490,6 +593,11 @@ class Interpreter(object):
                         attr=attr, loc=self.loc)
         self.current_block.append(sa)
 
+    def op_DELETE_ATTR(self, inst, target):
+        attr = self.code_names[inst.arg]
+        sa = ir.DelAttr(target=self.get(target), attr=attr, loc=self.loc)
+        self.current_block.append(sa)
+
     def op_LOAD_ATTR(self, inst, item, res):
         item = self.get(item)
         attr = self.code_names[inst.arg]
@@ -504,6 +612,7 @@ class Interpreter(object):
     def op_LOAD_GLOBAL(self, inst, res):
         name = self.code_names[inst.arg]
         value = self.get_global_value(name)
+        self.used_globals[name] = value
         gl = ir.Global(name, value, loc=self.loc)
         self.store(gl, res)
         self.constants[res] = value
@@ -511,15 +620,13 @@ class Interpreter(object):
     def op_LOAD_DEREF(self, inst, res):
         name = self.code_freevars[inst.arg]
         value = self.get_closure_value(inst.arg)
-        # closure values are treated like globals
-        gl = ir.Global(name, value, loc=self.loc)
+        gl = ir.FreeVar(inst.arg, name, value, loc=self.loc)
         self.store(gl, res)
 
     def op_SETUP_LOOP(self, inst):
         assert self.blocks[inst.offset] is self.current_block
         loop = ir.Loop(inst.offset, exit=(inst.next + inst.arg))
         self.syntax_blocks.append(loop)
-        self.syntax_info.append(loop)
 
     def op_CALL_FUNCTION(self, inst, func, args, kws, res):
         func = self.get(func)
@@ -552,10 +659,6 @@ class Interpreter(object):
         """
         assert inst.offset in self.blocks, "FOR_ITER must be block head"
 
-        # Mark this block as the loop condition
-        loop = self.syntax_blocks[-1]
-        loop.condition = self.current_block_offset
-
         # Emit code
         val = self.get(iterator)
 
@@ -574,12 +677,6 @@ class Interpreter(object):
                        loc=self.loc)
         self.current_block.append(br)
 
-        # Add event listener to mark the following blocks as loop body
-        def mark_as_body(offset, block):
-            loop.body.append(offset)
-
-        self._block_actions[loop] = mark_as_body
-
     def op_BINARY_SUBSCR(self, inst, target, index, res):
         index = self.get(index)
         target = self.get(target)
@@ -614,8 +711,9 @@ class Interpreter(object):
         self.store(expr, res)
 
     def op_STORE_MAP(self, inst, dct, key, value):
-        self.current_block.append(
-            ir.StoreMap(dct=dct, key=key, value=value, loc=self.loc))
+        stmt = ir.StoreMap(dct=self.get(dct), key=self.get(key),
+                           value=self.get(value), loc=self.loc)
+        self.current_block.append(stmt)
 
     def op_UNARY_NEGATIVE(self, inst, value, res):
         value = self.get(value)
@@ -643,6 +741,12 @@ class Interpreter(object):
         expr = ir.Expr.binop(op, lhs=lhs, rhs=rhs, loc=self.loc)
         self.store(expr, res)
 
+    def _inplace_binop(self, op, lhs, rhs, res):
+        lhs = self.get(lhs)
+        rhs = self.get(rhs)
+        expr = ir.Expr.inplace_binop(op, lhs=lhs, rhs=rhs, loc=self.loc)
+        self.store(expr, res)
+
     def op_BINARY_ADD(self, inst, lhs, rhs, res):
         self._binop('+', lhs, rhs, res)
 
@@ -682,8 +786,6 @@ class Interpreter(object):
     def op_BINARY_XOR(self, inst, lhs, rhs, res):
         self._binop('^', lhs, rhs, res)
 
-    _inplace_binop = _binop
-
     def op_INPLACE_ADD(self, inst, lhs, rhs, res):
         self._inplace_binop('+', lhs, rhs, res)
 
@@ -731,16 +833,12 @@ class Interpreter(object):
         jmp = ir.Jump(inst.get_jump_target(), loc=self.loc)
         self.current_block.append(jmp)
 
-    def op_POP_BLOCK(self, inst, delitems=()):
-        blk = self.syntax_blocks.pop()
-        for item in delitems:
-            delete = ir.Del(item, loc=self.loc)
-            self.current_block.append(delete)
-        if blk in self._block_actions:
-            del self._block_actions[blk]
+    def op_POP_BLOCK(self, inst):
+        self.syntax_blocks.pop()
 
-    def op_RETURN_VALUE(self, inst, retval):
-        ret = ir.Return(self.get(retval), loc=self.loc)
+    def op_RETURN_VALUE(self, inst, retval, castval):
+        self.store(ir.Expr.cast(self.get(retval), loc=self.loc), castval)
+        ret = ir.Return(self.get(castval), loc=self.loc)
         self.current_block.append(ret)
 
     def op_COMPARE_OP(self, inst, lhs, rhs, res):
@@ -763,8 +861,6 @@ class Interpreter(object):
         bra = ir.Branch(cond=self.get(pred), truebr=truebr, falsebr=falsebr,
                         loc=self.loc)
         self.current_block.append(bra)
-        # In a while loop?
-        self._determine_while_condition((truebr, falsebr))
 
     def op_JUMP_IF_FALSE(self, inst, pred):
         self._op_JUMP_IF(inst, pred=pred, iftrue=False)
@@ -787,33 +883,3 @@ class Interpreter(object):
     def op_RAISE_VARARGS(self, inst, exc):
         stmt = ir.Raise(exception=self.constants[exc], loc=self.loc)
         self.current_block.append(stmt)
-
-    def _determine_while_condition(self, branches):
-        assert branches
-        # There is a active syntax block
-        if not self.syntax_blocks:
-            return
-            # TOS is a Loop instance
-        loop = self.syntax_blocks[-1]
-        if not isinstance(loop, ir.Loop):
-            return
-            # Its condition is not defined
-        if loop.condition is not None:
-            return
-            # One of the branches goes to a POP_BLOCK
-        for br in branches:
-            if self.block_constains_opname(br, 'POP_BLOCK'):
-                break
-        else:
-            return
-            # Which is the exit of the loop
-        if br not in self.cfa.blocks[loop.exit].incoming_jumps:
-            return
-
-        # Therefore, current block is a while loop condition
-        loop.condition = self.current_block_offset
-        # Add event listener to mark the following blocks as loop body
-        def mark_as_body(offset, block):
-            loop.body.append(offset)
-
-        self._block_actions[loop] = mark_as_body
