@@ -4,16 +4,12 @@ from collections import defaultdict
 import sys
 from types import ModuleType
 
-from llvm.core import Type, Builder, Module
-import llvm.core as lc
+from llvmlite.llvmpy.core import Type, Builder, Module
+import llvmlite.llvmpy.core as lc
+import llvmlite.binding as ll
 
-from numba import ir, types, cgutils, utils, config, cffi_support, typing
-
-
-try:
-    import builtins
-except ImportError:
-    import __builtin__ as builtins
+from numba import (_dynfunc, errcode, ir, types, cgutils, utils, config,
+                   cffi_support, typing)
 
 
 class LoweringError(Exception):
@@ -36,26 +32,20 @@ def default_mangler(name, argtypes):
 _dynamic_modname = '<dynamic>'
 _dynamic_module = ModuleType(_dynamic_modname)
 
-# Issue #475: locals() is unsupported as calling it naively would give
-# out wrong results.
-_unsupported_builtins = set([locals])
-
 
 class FunctionDescriptor(object):
-    __slots__ = ('native', 'modname', 'qualname', 'doc', 'blocks', 'typemap',
+    __slots__ = ('native', 'modname', 'qualname', 'doc', 'typemap',
                  'calltypes', 'args', 'kws', 'restype', 'argtypes',
-                 'mangled_name', 'unique_name', 'globals')
+                 'mangled_name', 'unique_name')
 
-    def __init__(self, native, modname, qualname, unique_name, doc, blocks,
+    def __init__(self, native, modname, qualname, unique_name, doc,
                  typemap, restype, calltypes, args, kws, mangler=None,
-                 argtypes=None, globals=None):
+                 argtypes=None):
         self.native = native
         self.modname = modname
         self.qualname = qualname
         self.unique_name = unique_name
-        self.globals = globals if globals is not None else {}
         self.doc = doc
-        self.blocks = blocks
         self.typemap = typemap
         self.calltypes = calltypes
         self.args = args
@@ -64,7 +54,13 @@ class FunctionDescriptor(object):
         # Argument types
         self.argtypes = argtypes or [self.typemap[a] for a in args]
         mangler = default_mangler if mangler is None else mangler
-        self.mangled_name = mangler(self.qualname, self.argtypes)
+        # XXX The mangled name should really be unique but this formula
+        # doesn't guarantee it entirely.
+        if self.modname:
+            self.mangled_name = mangler('%s.%s' % (self.modname, self.qualname),
+                                        self.argtypes)
+        else:
+            self.mangled_name = mangler(self.qualname, self.argtypes)
 
     def lookup_module(self):
         """
@@ -77,6 +73,14 @@ class FunctionDescriptor(object):
         else:
             return sys.modules[self.modname]
 
+    @property
+    def llvm_func_name(self):
+        return self.mangled_name
+
+    @property
+    def llvm_cpython_wrapper_name(self):
+        return 'wrapper.' + self.mangled_name
+
     def __repr__(self):
         return "<function descriptor %r>" % (self.unique_name)
 
@@ -90,7 +94,7 @@ class FunctionDescriptor(object):
         ``unique_name`` must be a unique name in ``pymod``.
         """
         func = interp.bytecode.func
-        qualname = getattr(func, '__qualname__', interp.bytecode.func_name)
+        qualname = interp.bytecode.func_qualname
         modname = func.__module__
         doc = func.__doc__ or ''
         args = interp.argspec.args
@@ -107,17 +111,16 @@ class FunctionDescriptor(object):
             # TODO avoid unnecessary recompilation of the same function
             unique_name = "%s$%d" % (qualname, func.__code__.co_firstlineno)
 
-        return qualname, unique_name, modname, doc, args, kws, func.__globals__
+        return qualname, unique_name, modname, doc, args, kws
 
     @classmethod
     def _from_python_function(cls, interp, typemap, restype, calltypes,
                               native, mangler=None):
-        (qualname, unique_name, modname, doc, args, kws, func_globals
+        (qualname, unique_name, modname, doc, args, kws,
          )= cls._get_function_info(interp)
-        sortedblocks = utils.SortedMap(utils.dict_iteritems(interp.blocks))
         self = cls(native, modname, qualname, unique_name, doc,
-                   sortedblocks, typemap, restype, calltypes,
-                   args, kws, mangler=mangler, globals=func_globals)
+                   typemap, restype, calltypes,
+                   args, kws, mangler=mangler)
         return self
 
 
@@ -157,35 +160,41 @@ class ExternalFunctionDescriptor(FunctionDescriptor):
         args = ["arg%d" % i for i in range(len(argtypes))]
         super(ExternalFunctionDescriptor, self).__init__(native=True,
                 modname=None, qualname=name, unique_name=name, doc='',
-                blocks=None, typemap=None, restype=restype, calltypes=None,
-                args=args, kws=None, mangler=lambda a, x: a, argtypes=argtypes,
-                globals={})
+                typemap=None, restype=restype, calltypes=None,
+                args=args, kws=None, mangler=lambda a, x: a,
+                argtypes=argtypes)
 
 
 class BaseLower(object):
     """
     Lower IR to LLVM
     """
-    def __init__(self, context, fndesc):
+    def __init__(self, context, library, fndesc, interp):
         self.context = context
+        self.library = library
         self.fndesc = fndesc
-        # Initialize LLVM
-        self.module = Module.new("module.%s" % self.fndesc.unique_name)
+        self.blocks = utils.SortedMap(utils.iteritems(interp.blocks))
 
-        # Install metadata
-        md_pymod = cgutils.MetadataKeyStore(self.module, "python.module")
-        md_pymod.set(fndesc.modname)
+        # Initialize LLVM
+        self.module = self.library.create_ir_module(self.fndesc.unique_name)
+
+        # Python execution environment (will be available to the compiled
+        # function).
+        self.env = _dynfunc.Environment(
+            globals=self.fndesc.lookup_module().__dict__)
+
+        # Mapping of error codes to exception classes or instances
+        self.exceptions = {}
 
         # Setup function
         self.function = context.declare_function(self.module, fndesc)
         self.entry_block = self.function.append_basic_block('entry')
         self.builder = Builder.new(self.entry_block)
-        # self.builder = cgutils.VerboseProxy(self.builder)
 
         # Internal states
         self.blkmap = {}
         self.varmap = {}
-        self.firstblk = min(self.fndesc.blocks.keys())
+        self.firstblk = min(self.blocks.keys())
         self.loc = -1
 
         # Subclass initialization
@@ -194,12 +203,23 @@ class BaseLower(object):
     def init(self):
         pass
 
-    def post_lower(self):
-        """Called after all blocks are lowered
+    def pre_lower(self):
+        """
+        Called before lowering all blocks.
         """
-        pass
 
-    def lower(self):
+    def post_lower(self):
+        """
+        Called after all blocks are lowered
+        """
+
+    def add_exception(self, exc):
+        assert issubclass(exc, BaseException), exc
+        excid = len(self.exceptions) + errcode.ERROR_COUNT
+        self.exceptions[excid] = exc
+        return excid
+
+    def lower(self, create_wrapper=True):
         # Init argument variables
         fnargs = self.context.get_arguments(self.function)
         for ak, av in zip(self.fndesc.args, fnargs):
@@ -207,35 +227,43 @@ class BaseLower(object):
             av = self.context.get_argument_value(self.builder, at, av)
             av = self.init_argument(av)
             self.storevar(av, ak)
+
         # Init blocks
-        for offset in self.fndesc.blocks:
-            bname = "B%d" % offset
+        for offset in self.blocks:
+            bname = "B%s" % offset
             self.blkmap[offset] = self.function.append_basic_block(bname)
+
+        self.pre_lower()
+        # pre_lower() may have changed the current basic block
+        entry_block_tail = self.builder.basic_block
+
         # Lower all blocks
-        for offset, block in self.fndesc.blocks.items():
+        for offset, block in self.blocks.items():
             bb = self.blkmap[offset]
             self.builder.position_at_end(bb)
             self.lower_block(block)
 
         self.post_lower()
-        # Close entry block
-        self.builder.position_at_end(self.entry_block)
+
+        # Close tail of entry block
+        self.builder.position_at_end(entry_block_tail)
         self.builder.branch(self.blkmap[self.firstblk])
 
+        # Run target specific post lowering transformation
+        self.context.post_lowering(self.function)
+
         if config.DUMP_LLVM:
             print(("LLVM DUMP %s" % self.fndesc).center(80, '-'))
             print(self.module)
             print('=' * 80)
-        self.module.verify()
-        # Run function-level optimize to reduce memory usage and improve
-        # module-level optimization
-        self.context.optimize_function(self.function)
 
-        if config.NUMBA_DUMP_FUNC_OPT:
-            print(("LLVM FUNCTION OPTIMIZED DUMP %s" %
-                   self.fndesc).center(80, '-'))
-            print(self.module)
-            print('=' * 80)
+        # Materialize LLVM Module
+        self.library.add_ir_module(self.module)
+
+        # Create CPython wrapper
+        if create_wrapper:
+            self.context.create_cpython_wrapper(self.library, self.fndesc,
+                                                self.exceptions)
 
     def init_argument(self, arg):
         return arg
@@ -283,12 +311,9 @@ class Lower(BaseLower):
             oty = self.typeof(inst.value.name)
             ty = self.fndesc.restype
             if isinstance(ty, types.Optional):
-                if oty == types.none:
-                    self.context.return_native_none(self.builder)
-                    return
-                else:
-                    ty = ty.type
-
+                # If returning an optional type
+                self.context.return_optional_value(self.builder, ty, oty, val)
+                return
             if ty != oty:
                 val = self.context.cast(self.builder, val, oty, ty)
             retval = self.context.get_return_value(self.builder, ty, val)
@@ -308,7 +333,12 @@ class Lower(BaseLower):
             impl = self.context.get_function('setitem', signature)
 
             # Convert argument to match
-            assert targetty == signature.args[0]
+            if isinstance(targetty, types.Optional):
+                target = self.context.cast(self.builder, target, targetty,
+                                           targetty.type)
+            else:
+                assert targetty == signature.args[0]
+
             index = self.context.cast(self.builder, index, indexty,
                                       signature.args[1])
             value = self.context.cast(self.builder, value, valuety,
@@ -337,7 +367,7 @@ class Lower(BaseLower):
             return impl(self.builder, (target, value))
 
         elif isinstance(inst, ir.Raise):
-            excid = self.context.add_exception(inst.exception)
+            excid = self.add_exception(inst.exception)
             self.context.return_user_exc(self.builder, excid)
 
         else:
@@ -357,7 +387,8 @@ class Lower(BaseLower):
             oty = self.typeof(value.name)
             return self.context.cast(self.builder, val, oty, ty)
 
-        elif isinstance(value, ir.Global):
+        # In nopython mode, closure vars are frozen like globals
+        elif isinstance(value, (ir.Global, ir.FreeVar)):
             if (isinstance(ty, types.Dummy) or
                     isinstance(ty, types.Module) or
                     isinstance(ty, types.Function) or
@@ -375,24 +406,32 @@ class Lower(BaseLower):
         else:
             raise NotImplementedError(type(value), value)
 
+    def lower_binop(self, resty, expr):
+        lhs = expr.lhs
+        rhs = expr.rhs
+        lty = self.typeof(lhs.name)
+        rty = self.typeof(rhs.name)
+        lhs = self.loadvar(lhs.name)
+        rhs = self.loadvar(rhs.name)
+        # Get function
+        signature = self.fndesc.calltypes[expr]
+        impl = self.context.get_function(expr.fn, signature)
+        # Convert argument to match
+        lhs = self.context.cast(self.builder, lhs, lty, signature.args[0])
+        rhs = self.context.cast(self.builder, rhs, rty, signature.args[1])
+        res = impl(self.builder, (lhs, rhs))
+        return self.context.cast(self.builder, res, signature.return_type,
+                                 resty)
+
     def lower_expr(self, resty, expr):
         if expr.op == 'binop':
-            lhs = expr.lhs
-            rhs = expr.rhs
-            lty = self.typeof(lhs.name)
-            rty = self.typeof(rhs.name)
-            lhs = self.loadvar(lhs.name)
-            rhs = self.loadvar(rhs.name)
-            # Get function
-            signature = self.fndesc.calltypes[expr]
-            impl = self.context.get_function(expr.fn, signature)
-            # Convert argument to match
-            lhs = self.context.cast(self.builder, lhs, lty, signature.args[0])
-            rhs = self.context.cast(self.builder, rhs, rty, signature.args[1])
-            res = impl(self.builder, (lhs, rhs))
-            return self.context.cast(self.builder, res, signature.return_type,
-                                     resty)
-
+            return self.lower_binop(resty, expr)
+        elif expr.op == 'inplace_binop':
+            lty = self.typeof(expr.lhs.name)
+            if not lty.mutable:
+                # inplace operators on non-mutable types reuse the same
+                # definition as the corresponding copying operators.
+                return self.lower_binop(resty, expr)
         elif expr.op == 'unary':
             val = self.loadvar(expr.value.name)
             typ = self.typeof(expr.value.name)
@@ -445,10 +484,17 @@ class Lower(BaseLower):
             else:
                 # Normal function resolution
                 impl = self.context.get_function(fnty, signature)
+                if signature.recvr:
+                    # The "self" object is passed as the function object
+                    # for bounded function
+                    the_self = self.loadvar(expr.func.name)
+                    # Prepend the self reference
+                    castvals = [the_self] + castvals
+
                 res = impl(self.builder, castvals)
                 libs = getattr(impl, "libs", ())
-                if libs:
-                    self.context.add_libs(libs)
+                for lib in libs:
+                    self.library.add_linking_library(lib)
             return self.context.cast(self.builder, res, signature.return_type,
                                      resty)
 
@@ -495,7 +541,7 @@ class Lower(BaseLower):
             iternext_impl = self.context.get_function('iternext',
                                                       iternext_sig)
             iterobj = getiter_impl(self.builder, (val,))
-            excid = self.context.add_exception(ValueError)
+            excid = self.add_exception(ValueError)
             # We call iternext() as many times as desired (`expr.count`).
             for i in range(expr.count):
                 pair = iternext_impl(self.builder, (iterobj,))
@@ -521,12 +567,19 @@ class Lower(BaseLower):
         elif expr.op == "getattr":
             val = self.loadvar(expr.value.name)
             ty = self.typeof(expr.value.name)
-            impl = self.context.get_attribute(val, ty, expr.attr)
-            if impl is None:
-                # ignore the attribute
-                res = self.context.get_dummy_value()
+
+            if isinstance(resty, types.BoundFunction):
+                # if we are getting out a method, assume we have typed this
+                # properly and just build a bound function object
+                res = self.context.get_bound_function(self.builder, val, ty)
             else:
-                res = impl(self.context, self.builder, ty, val, expr.attr)
+                impl = self.context.get_attribute(val, ty, expr.attr)
+
+                if impl is None:
+                    # ignore the attribute
+                    res = self.context.get_dummy_value()
+                else:
+                    res = impl(self.context, self.builder, ty, val, expr.attr)
             return res
 
         elif expr.op == "static_getitem":
@@ -573,6 +626,12 @@ class Lower(BaseLower):
                 tup = self.builder.insert_value(tup, castvals[i], i)
             return tup
 
+        elif expr.op == "cast":
+            val = self.loadvar(expr.value.name)
+            ty = self.typeof(expr.value.name)
+            castval = self.context.cast(self.builder, val, ty, resty)
+            return castval
+
         raise NotImplementedError(expr)
 
     def getvar(self, name):
@@ -595,461 +654,4 @@ class Lower(BaseLower):
         return self.alloca_lltype(name, lltype)
 
     def alloca_lltype(self, name, lltype):
-        bb = self.builder.basic_block
-        self.builder.position_at_end(self.entry_block)
-        ptr = self.builder.alloca(lltype, name=name)
-        self.builder.position_at_end(bb)
-        return ptr
-
-
-PYTHON_OPMAP = {
-     '+': "number_add",
-     '-': "number_subtract",
-     '*': "number_multiply",
-    '/?': "number_divide",
-     '/': "number_truedivide",
-    '//': "number_floordivide",
-     '%': "number_remainder",
-    '**': "number_power",
-    '<<': "number_lshift",
-    '>>': "number_rshift",
-     '&': "number_and",
-     '|': "number_or",
-     '^': "number_xor",
-}
-
-
-class PyLower(BaseLower):
-    def init(self):
-        self.pyapi = self.context.get_python_api(self.builder)
-
-        # Add error handling block
-        self.ehblock = self.function.append_basic_block('error')
-
-    def post_lower(self):
-        with cgutils.goto_block(self.builder, self.ehblock):
-            self.cleanup()
-            self.context.return_exc(self.builder)
-
-    def init_argument(self, arg):
-        self.incref(arg)
-        return arg
-
-    def lower_inst(self, inst):
-        if isinstance(inst, ir.Assign):
-            value = self.lower_assign(inst)
-            self.storevar(value, inst.target.name)
-
-        elif isinstance(inst, ir.SetItem):
-            target = self.loadvar(inst.target.name)
-            index = self.loadvar(inst.index.name)
-            value = self.loadvar(inst.value.name)
-            ok = self.pyapi.object_setitem(target, index, value)
-            self.check_int_status(ok)
-
-        elif isinstance(inst, ir.StoreMap):
-            dct = self.loadvar(inst.dct)
-            key = self.loadvar(inst.key)
-            value = self.loadvar(inst.value)
-            ok = self.pyapi.dict_setitem(dct, key, value)
-            self.check_int_status(ok)
-
-        elif isinstance(inst, ir.Return):
-            retval = self.loadvar(inst.value.name)
-            self.incref(retval)
-            self.cleanup()
-            self.context.return_value(self.builder, retval)
-
-        elif isinstance(inst, ir.Branch):
-            cond = self.loadvar(inst.cond.name)
-            if cond.type == Type.int(1):
-                istrue = cond
-            else:
-                istrue = self.pyapi.object_istrue(cond)
-            zero = lc.Constant.null(istrue.type)
-            pred = self.builder.icmp(lc.ICMP_NE, istrue, zero)
-            tr = self.blkmap[inst.truebr]
-            fl = self.blkmap[inst.falsebr]
-            self.builder.cbranch(pred, tr, fl)
-
-        elif isinstance(inst, ir.Jump):
-            target = self.blkmap[inst.target]
-            self.builder.branch(target)
-
-        elif isinstance(inst, ir.Del):
-            self.delvar(inst.value)
-
-        elif isinstance(inst, ir.Raise):
-            self.pyapi.raise_exception(inst.exception, inst.exception)
-            self.return_exception_raised()
-
-        else:
-            raise NotImplementedError(type(inst), inst)
-
-    def lower_assign(self, inst):
-        """
-        The returned object must have a new reference
-        """
-        value = inst.value
-        if isinstance(value, ir.Const):
-            return self.lower_const(value.value)
-        elif isinstance(value, ir.Var):
-            val = self.loadvar(value.name)
-            self.incref(val)
-            return val
-        elif isinstance(value, ir.Expr):
-            return self.lower_expr(value)
-        elif isinstance(value, ir.Global):
-            return self.lower_global(value.name, value.value)
-        else:
-            raise NotImplementedError(type(value), value)
-
-    def lower_expr(self, expr):
-        if expr.op == 'binop':
-            lhs = self.loadvar(expr.lhs.name)
-            rhs = self.loadvar(expr.rhs.name)
-            if expr.fn in PYTHON_OPMAP:
-                fname = PYTHON_OPMAP[expr.fn]
-                fn = getattr(self.pyapi, fname)
-                res = fn(lhs, rhs)
-            else:
-                # Assume to be rich comparision
-                res = self.pyapi.object_richcompare(lhs, rhs, expr.fn)
-            self.check_error(res)
-            return res
-        elif expr.op == 'unary':
-            value = self.loadvar(expr.value.name)
-            if expr.fn == '-':
-                res = self.pyapi.number_negative(value)
-            elif expr.fn == '+':
-                res = self.pyapi.number_positive(value)
-            elif expr.fn == 'not':
-                res = self.pyapi.object_not(value)
-                self.check_int_status(res)
-
-                longval = self.builder.zext(res, self.pyapi.long)
-                res = self.pyapi.bool_from_long(longval)
-            elif expr.fn == '~':
-                res = self.pyapi.number_invert(value)
-            else:
-                raise NotImplementedError(expr)
-            self.check_error(res)
-            return res
-        elif expr.op == 'call':
-            argvals = [self.loadvar(a.name) for a in expr.args]
-            fn = self.loadvar(expr.func.name)
-            if not expr.kws:
-                # No keyword
-                ret = self.pyapi.call_function_objargs(fn, argvals)
-            else:
-                # Have Keywords
-                keyvalues = [(k, self.loadvar(v.name)) for k, v in expr.kws]
-                args = self.pyapi.tuple_pack(argvals)
-                kws = self.pyapi.dict_pack(keyvalues)
-                ret = self.pyapi.call(fn, args, kws)
-                self.decref(kws)
-                self.decref(args)
-            self.check_error(ret)
-            return ret
-        elif expr.op == 'getattr':
-            obj = self.loadvar(expr.value.name)
-            res = self.pyapi.object_getattr_string(obj, expr.attr)
-            self.check_error(res)
-            return res
-        elif expr.op == 'build_tuple':
-            items = [self.loadvar(it.name) for it in expr.items]
-            res = self.pyapi.tuple_pack(items)
-            self.check_error(res)
-            return res
-        elif expr.op == 'build_list':
-            items = [self.loadvar(it.name) for it in expr.items]
-            res = self.pyapi.list_pack(items)
-            self.check_error(res)
-            return res
-        elif expr.op == 'build_map':
-            res = self.pyapi.dict_new(expr.size)
-            self.check_error(res)
-            return res
-        elif expr.op == 'build_set':
-            items = [self.loadvar(it.name) for it in expr.items]
-            res = self.pyapi.set_new()
-            self.check_error(res)
-            for it in items:
-                ok = self.pyapi.set_add(res, it)
-                self.check_int_status(ok)
-            return res
-        elif expr.op == 'getiter':
-            obj = self.loadvar(expr.value.name)
-            res = self.pyapi.object_getiter(obj)
-            self.check_error(res)
-            return res
-        elif expr.op == 'iternext':
-            iterobj = self.loadvar(expr.value.name)
-            item = self.pyapi.iter_next(iterobj)
-            is_valid = cgutils.is_not_null(self.builder, item)
-            pair = self.pyapi.tuple_new(2)
-            with cgutils.ifelse(self.builder, is_valid) as (then, otherwise):
-                with then:
-                    self.pyapi.tuple_setitem(pair, 0, item)
-                with otherwise:
-                    self.check_occurred()
-                    # Make the tuple valid by inserting None as dummy
-                    # iteration "result" (it will be ignored).
-                    self.pyapi.tuple_setitem(pair, 0, self.pyapi.make_none())
-            self.pyapi.tuple_setitem(pair, 1, self.pyapi.bool_from_bool(is_valid))
-            return pair
-        elif expr.op == 'pair_first':
-            pair = self.loadvar(expr.value.name)
-            first = self.pyapi.tuple_getitem(pair, 0)
-            self.incref(first)
-            return first
-        elif expr.op == 'pair_second':
-            pair = self.loadvar(expr.value.name)
-            second = self.pyapi.tuple_getitem(pair, 1)
-            self.incref(second)
-            return second
-        elif expr.op == 'exhaust_iter':
-            iterobj = self.loadvar(expr.value.name)
-            tup = self.pyapi.sequence_tuple(iterobj)
-            self.check_error(tup)
-            # Check tuple size is as expected
-            tup_size = self.pyapi.tuple_size(tup)
-            expected_size = self.context.get_constant(types.intp, expr.count)
-            has_wrong_size = self.builder.icmp(lc.ICMP_NE,
-                                               tup_size, expected_size)
-            with cgutils.if_unlikely(self.builder, has_wrong_size):
-                excid = self.context.add_exception(ValueError)
-                self.context.return_user_exc(self.builder, excid)
-            return tup
-        elif expr.op == 'getitem':
-            value = self.loadvar(expr.value.name)
-            index = self.loadvar(expr.index.name)
-            res = self.pyapi.object_getitem(value, index)
-            self.check_error(res)
-            return res
-        elif expr.op == 'static_getitem':
-            value = self.loadvar(expr.value.name)
-            index = self.context.get_constant(types.intp, expr.index)
-            indexobj = self.pyapi.long_from_ssize_t(index)
-            self.check_error(indexobj)
-            res = self.pyapi.object_getitem(value, indexobj)
-            self.decref(indexobj)
-            self.check_error(res)
-            return res
-        elif expr.op == 'getslice':
-            target = self.loadvar(expr.target.name)
-            start = self.loadvar(expr.start.name)
-            stop = self.loadvar(expr.stop.name)
-
-            slicefn = self.get_builtin_obj("slice")
-            sliceobj = self.pyapi.call_function_objargs(slicefn, (start, stop))
-            self.decref(slicefn)
-            self.check_error(sliceobj)
-
-            res = self.pyapi.object_getitem(target, sliceobj)
-            self.check_error(res)
-
-            return res
-        else:
-            raise NotImplementedError(expr)
-
-    def lower_const(self, const):
-        if isinstance(const, str):
-            ret = self.pyapi.string_from_constant_string(const)
-            self.check_error(ret)
-            return ret
-        elif isinstance(const, complex):
-            real = self.context.get_constant(types.float64, const.real)
-            imag = self.context.get_constant(types.float64, const.imag)
-            ret = self.pyapi.complex_from_doubles(real, imag)
-            self.check_error(ret)
-            return ret
-        elif isinstance(const, float):
-            fval = self.context.get_constant(types.float64, const)
-            ret = self.pyapi.float_from_double(fval)
-            self.check_error(ret)
-            return ret
-        elif isinstance(const, utils.INT_TYPES):
-            if utils.bit_length(const) >= 64:
-                raise ValueError("Integer is too big to be lowered")
-            ival = self.context.get_constant(types.intp, const)
-            return self.pyapi.long_from_ssize_t(ival)
-        elif isinstance(const, tuple):
-            items = [self.lower_const(i) for i in const]
-            return self.pyapi.tuple_pack(items)
-        elif const is Ellipsis:
-            return self.get_builtin_obj("Ellipsis")
-        elif const is None:
-            return self.pyapi.make_none()
-        else:
-            raise NotImplementedError(type(const))
-
-    def lower_global(self, name, value):
-        """
-        1) Check global scope dictionary.
-        2) Check __builtins__.
-            2a) is it a dictionary (for non __main__ module)
-            2b) is it a module (for __main__ module)
-        """
-        moddict = self.pyapi.get_module_dict()
-        obj = self.pyapi.dict_getitem_string(moddict, name)
-        self.incref(obj)  # obj is borrowed
-
-        if value in _unsupported_builtins:
-            raise ForbiddenConstruct("builtins %s() is not supported"
-                                     % name, loc=self.loc)
-
-        if hasattr(builtins, name):
-            obj_is_null = self.is_null(obj)
-            bbelse = self.builder.basic_block
-
-            with cgutils.ifthen(self.builder, obj_is_null):
-                mod = self.pyapi.dict_getitem_string(moddict, "__builtins__")
-                builtin = self.builtin_lookup(mod, name)
-                bbif = self.builder.basic_block
-
-            retval = self.builder.phi(self.pyapi.pyobj)
-            retval.add_incoming(obj, bbelse)
-            retval.add_incoming(builtin, bbif)
-
-        else:
-            retval = obj
-            with cgutils.if_unlikely(self.builder, self.is_null(retval)):
-                self.pyapi.raise_missing_global_error(name)
-                self.return_exception_raised()
-
-        self.incref(retval)
-        return retval
-
-    # -------------------------------------------------------------------------
-
-    def get_builtin_obj(self, name):
-        moddict = self.pyapi.get_module_dict()
-        mod = self.pyapi.dict_getitem_string(moddict, "__builtins__")
-        return self.builtin_lookup(mod, name)
-
-    def builtin_lookup(self, mod, name):
-        """
-        Args
-        ----
-        mod:
-            The __builtins__ dictionary or module
-        name: str
-            The object to lookup
-        """
-        fromdict = self.pyapi.dict_getitem_string(mod, name)
-        self.incref(fromdict)       # fromdict is borrowed
-        bbifdict = self.builder.basic_block
-
-        with cgutils.if_unlikely(self.builder, self.is_null(fromdict)):
-            # This happen if we are using the __main__ module
-            frommod = self.pyapi.object_getattr_string(mod, name)
-
-            with cgutils.if_unlikely(self.builder, self.is_null(frommod)):
-                self.pyapi.raise_missing_global_error(name)
-                self.return_exception_raised()
-
-            bbifmod = self.builder.basic_block
-
-        builtin = self.builder.phi(self.pyapi.pyobj)
-        builtin.add_incoming(fromdict, bbifdict)
-        builtin.add_incoming(frommod, bbifmod)
-
-        return builtin
-
-    def check_occurred(self):
-        err_occurred = cgutils.is_not_null(self.builder,
-                                           self.pyapi.err_occurred())
-
-        with cgutils.if_unlikely(self.builder, err_occurred):
-            self.return_exception_raised()
-
-    def check_error(self, obj):
-        with cgutils.if_unlikely(self.builder, self.is_null(obj)):
-            self.return_exception_raised()
-
-        return obj
-
-    def check_int_status(self, num, ok_value=0):
-        """
-        Raise an exception if *num* is smaller than *ok_value*.
-        """
-        ok = lc.Constant.int(num.type, ok_value)
-        pred = self.builder.icmp(lc.ICMP_SLT, num, ok)
-        with cgutils.if_unlikely(self.builder, pred):
-            self.return_exception_raised()
-
-    def is_null(self, obj):
-        return cgutils.is_null(self.builder, obj)
-
-    def return_exception_raised(self):
-        self.builder.branch(self.ehblock)
-
-    def return_error_occurred(self):
-        self.cleanup()
-        self.context.return_exc(self.builder)
-
-    def getvar(self, name, ltype=None):
-        if name not in self.varmap:
-            self.varmap[name] = self.alloca(name, ltype=ltype)
-        return self.varmap[name]
-
-    def loadvar(self, name):
-        ptr = self.getvar(name)
-        return self.builder.load(ptr)
-
-    def delvar(self, name):
-        """
-        Delete the variable slot with the given name. This will decref
-        the corresponding Python object.
-        """
-        ptr = self.varmap.pop(name)
-        self.decref(ptr)
-
-    def storevar(self, value, name):
-        """
-        Stores a llvm value and allocate stack slot if necessary.
-        The llvm value can be of arbitrary type.
-        """
-        ptr = self.getvar(name, ltype=value.type)
-        old = self.builder.load(ptr)
-        assert value.type == ptr.type.pointee, (str(value.type),
-                                                str(ptr.type.pointee))
-        self.builder.store(value, ptr)
-        # Safe to call decref even on non python object
-        self.decref(old)
-
-    def cleanup(self):
-        for var in utils.dict_itervalues(self.varmap):
-            self.decref(self.builder.load(var))
-
-    def alloca(self, name, ltype=None):
-        """
-        Allocate a stack slot and initialize it to NULL.
-        The default is to allocate a pyobject pointer.
-        Use ``ltype`` to override.
-        """
-        if ltype is None:
-            ltype = self.context.get_value_type(types.pyobject)
-        bb = self.builder.basic_block
-        self.builder.position_at_end(self.entry_block)
-        ptr = self.builder.alloca(ltype, name=name)
-        self.builder.store(cgutils.get_null_value(ltype), ptr)
-        self.builder.position_at_end(bb)
-        return ptr
-
-    def incref(self, value):
-        self.pyapi.incref(value)
-
-    def decref(self, value):
-        """
-        This is allow to be called on non pyobject pointer, in which case
-        no code is inserted.
-        """
-        lpyobj = self.context.get_value_type(types.pyobject)
-
-        if value.type.kind == lc.TYPE_POINTER:
-            if value.type != lpyobj:
-                pass
-            else:
-                self.pyapi.decref(value)
+        return cgutils.alloca_once(self.builder, lltype, name=name)
