@@ -1,7 +1,9 @@
 from __future__ import print_function, division, absolute_import
 import numpy as np
-from llvm.core import (Type, Builder, inline_function, LINKAGE_INTERNAL,
-                       ICMP_EQ, Constant)
+from llvmlite.llvmpy.core import (Type, Builder, LINKAGE_INTERNAL,
+                                  ICMP_EQ, Constant)
+import llvmlite.llvmpy.core as lc
+from llvmlite import binding as ll
 
 from numba import types, cgutils, config
 from numba import _dynfunc
@@ -33,6 +35,33 @@ def _build_ufunc_loop_body(load, store, context, func, builder, arrays, out,
     return status.code
 
 
+def _build_ufunc_loop_body_objmode(load, store, context, func, builder,
+                                   arrays, out, offsets, store_offset,
+                                   signature, env):
+    elems = load()
+
+    # Compute
+    _objargs = [types.pyobject] * len(signature.args)
+    status, retval = context.call_function(builder, func, types.pyobject,
+                                           _objargs, elems, env=env)
+
+    # Ignoring error status and store result
+    # Store
+    if out.byref:
+        retval = builder.load(retval)
+
+    store(retval)
+
+    # increment indices
+    for off, ary in zip(offsets, arrays):
+        builder.store(builder.add(builder.load(off), ary.step), off)
+
+    builder.store(builder.add(builder.load(store_offset), out.step),
+                  store_offset)
+
+    return status.code
+
+
 def build_slow_loop_body(context, func, builder, arrays, out, offsets,
                          store_offset, signature):
     def load():
@@ -47,6 +76,28 @@ def build_slow_loop_body(context, func, builder, arrays, out, offsets,
                                   out, offsets, store_offset, signature)
 
 
+def build_obj_loop_body(context, func, builder, arrays, out, offsets,
+                        store_offset, signature, pyapi, env):
+    def load():
+        # Load
+        elems = [ary.load_direct(builder.load(off))
+                 for off, ary in zip(offsets, arrays)]
+        # Box
+        elems = [pyapi.from_native_value(v, t)
+                 for v, t in zip(elems, signature.args)]
+        return elems
+
+    def store(retval):
+        # Unbox
+        retval = pyapi.to_native_value(retval, signature.return_type)
+        # Store
+        out.store_direct(retval, builder.load(store_offset))
+
+    return _build_ufunc_loop_body_objmode(load, store, context, func, builder,
+                                          arrays, out, offsets, store_offset,
+                                          signature, env)
+
+
 def build_fast_loop_body(context, func, builder, arrays, out, offsets,
                          store_offset, signature, ind):
     def load():
@@ -61,7 +112,7 @@ def build_fast_loop_body(context, func, builder, arrays, out, offsets,
                                   out, offsets, store_offset, signature)
 
 
-def build_ufunc_wrapper(context, func, signature):
+def build_ufunc_wrapper(library, context, func, signature, objmode, env):
     """
     Wrap the scalar function with a loop that iterates over the arguments
     """
@@ -76,7 +127,19 @@ def build_ufunc_wrapper(context, func, signature):
     fnty = Type.function(Type.void(), [byte_ptr_ptr_t, intp_ptr_t,
                                        intp_ptr_t, byte_ptr_t])
 
-    wrapper = module.add_function(fnty, "__ufunc__." + func.name)
+    wrapper_module = library.create_ir_module('')
+    if objmode:
+        func_type = context.get_function_type2(
+            types.pyobject, [types.pyobject] * len(signature.args))
+    else:
+        func_type = context.get_function_type2(signature.return_type,
+                                               signature.args)
+    oldfunc = func
+    func = wrapper_module.add_function(func_type,
+                                       name=func.name)
+    func.attributes.add("alwaysinline")
+
+    wrapper = wrapper_module.add_function(fnty, "__ufunc__." + func.name)
     arg_args, arg_dims, arg_steps, arg_data = wrapper.args
     arg_args.name = "args"
     arg_dims.name = "dims"
@@ -115,38 +178,48 @@ def build_ufunc_wrapper(context, func, signature):
     for ary in arrays:
         unit_strided = builder.and_(unit_strided, ary.is_unit_strided)
 
-    with cgutils.ifelse(builder, unit_strided) as (is_unit_strided,
-                                                   is_strided):
+    if objmode:
+    # General loop
+        pyapi = context.get_python_api(builder)
+        gil = pyapi.gil_ensure()
+        with cgutils.for_range(builder, loopcount, intp=intp_t):
+            slowloop = build_obj_loop_body(context, func, builder,
+                                           arrays, out, offsets,
+                                           store_offset, signature,
+                                           pyapi, env)
+        pyapi.gil_release(gil)
+        builder.ret_void()
 
-        with is_unit_strided:
-            with cgutils.for_range(builder, loopcount, intp=intp_t) as ind:
-                fastloop = build_fast_loop_body(context, func, builder,
-                                                arrays, out, offsets,
-                                                store_offset, signature, ind)
-            builder.ret_void()
+    else:
 
-        with is_strided:
-            # General loop
-            with cgutils.for_range(builder, loopcount, intp=intp_t):
-                slowloop = build_slow_loop_body(context, func, builder,
-                                                arrays, out, offsets,
-                                                store_offset, signature)
+        with cgutils.ifelse(builder, unit_strided) as (is_unit_strided,
+                                                       is_strided):
 
-            builder.ret_void()
+            with is_unit_strided:
+                with cgutils.for_range(builder, loopcount, intp=intp_t) as ind:
+                    fastloop = build_fast_loop_body(context, func, builder,
+                                                    arrays, out, offsets,
+                                                    store_offset, signature,
+                                                    ind)
+                builder.ret_void()
 
-    builder.ret_void()
+            with is_strided:
+                # General loop
+                with cgutils.for_range(builder, loopcount, intp=intp_t):
+                    slowloop = build_slow_loop_body(context, func, builder,
+                                                    arrays, out, offsets,
+                                                    store_offset, signature)
+
+                builder.ret_void()
+
+        builder.ret_void()
     del builder
 
-    # Set core function to internal so that it is not generated
-    func.linkage = LINKAGE_INTERNAL
-    # Force inline of code function
-    inline_function(slowloop)
-    inline_function(fastloop)
-    # Run optimizer
-    context.optimize(module)
 
-    if config.DUMP_OPTIMIZED:
-        print(module)
+    # Run optimizer
+    library.add_ir_module(wrapper_module)
+    wrapper = library.get_function(wrapper.name)
+    oldfunc.linkage = LINKAGE_INTERNAL
 
     return wrapper
 
@@ -190,7 +263,7 @@ class UArrayArg(object):
 
     def store_direct(self, value, offset):
         ptr = cgutils.pointer_add(self.builder, self.data, offset)
-        assert ptr.type.pointee == value.type
+        assert ptr.type.pointee == value.type, (ptr.type, value.type)
         self.builder.store(value, ptr)
 
     def store_aligned(self, value, ind):
@@ -199,7 +272,8 @@ class UArrayArg(object):
 
 
 class _GufuncWrapper(object):
-    def __init__(self, context, func, signature, sin, sout, fndesc):
+    def __init__(self, library, context, func, signature, sin, sout, fndesc):
+        self.library = library
         self.context = context
         self.func = func
         self.signature = signature
@@ -221,7 +295,12 @@ class _GufuncWrapper(object):
         fnty = Type.function(Type.void(), [byte_ptr_ptr_t, intp_ptr_t,
                                            intp_ptr_t, byte_ptr_t])
 
-        wrapper = module.add_function(fnty, "__gufunc__." + self.func.name)
+        wrapper_module = self.library.create_ir_module('')
+        func_type = self.context.get_function_type(self.fndesc)
+        func = wrapper_module.add_function(func_type, name=self.func.name)
+        func.attributes.add("alwaysinline")
+        wrapper = wrapper_module.add_function(fnty,
+                                              "__gufunc__." + self.func.name)
         arg_args, arg_dims, arg_steps, arg_data = wrapper.args
         arg_args.name = "args"
         arg_dims.name = "dims"
@@ -269,7 +348,7 @@ class _GufuncWrapper(object):
         # Loop
         with cgutils.for_range(builder, loopcount, intp=intp_t) as ind:
             args = [a.array_value for a in arrays]
-            innercall, error = self.gen_loop_body(builder, args)
+            innercall, error = self.gen_loop_body(builder, func, args)
             # If error, escape
             cgutils.cbranch_or_continue(builder, error, bbreturn)
 
@@ -284,24 +363,19 @@ class _GufuncWrapper(object):
 
         builder.ret_void()
 
-        module.verify()
+        self.library.add_ir_module(wrapper_module)
+        wrapper = self.library.get_function(wrapper.name)
+
         # Set core function to internal so that it is not generated
         self.func.linkage = LINKAGE_INTERNAL
-        # Force inline of code function
-        inline_function(innercall)
-        # Run optimizer
-        self.context.optimize(module)
 
-        if config.DUMP_OPTIMIZED:
-            print(module)
-
-        wrapper.verify()
         return wrapper, self.env
 
-    def gen_loop_body(self, builder, args):
-        status, retval = self.context.call_function(builder, self.func,
+    def gen_loop_body(self, builder, func, args):
+        status, retval = self.context.call_function(builder, func,
                                                     self.signature.return_type,
                                                     self.signature.args, args)
+
         innercall = status.code
         error = status.err
         return innercall, error
@@ -314,9 +388,9 @@ class _GufuncWrapper(object):
 
 
 class _GufuncObjectWrapper(_GufuncWrapper):
-    def gen_loop_body(self, builder, args):
+    def gen_loop_body(self, builder, func, args):
         innercall, error = _prepare_call_to_object_mode(self.context,
-                                                        builder, self.func,
+                                                        builder, func,
                                                         self.signature,
                                                         args, env=self.envptr)
         return innercall, error
@@ -339,11 +413,11 @@ class _GufuncObjectWrapper(_GufuncWrapper):
         self.pyapi.gil_release(self.gil)
 
 
-def build_gufunc_wrapper(context, func, signature, sin, sout, fndesc):
+def build_gufunc_wrapper(library, context, func, signature, sin, sout, fndesc):
     wrapcls = (_GufuncObjectWrapper
                if signature.return_type == types.pyobject
                else _GufuncWrapper)
-    return wrapcls(context, func, signature, sin, sout, fndesc).build()
+    return wrapcls(library, context, func, signature, sin, sout, fndesc).build()
 
 
 def _prepare_call_to_object_mode(context, builder, func, signature, args,
@@ -372,7 +446,7 @@ def _prepare_call_to_object_mode(context, builder, func, signature, args,
                                     ll_intp_ptr, ll_voidptr,
                                     ll_int, ll_int])
 
-    fn_array_new = mod.get_or_insert_function(fnty, name="NumbaNDArrayNew")
+    fn_array_new = mod.get_or_insert_function(fnty, name="numba_ndarray_new")
 
     # Convert each llarray into pyobject
     error_pointer = cgutils.alloca_once(builder, Type.int(1), name='error')
