@@ -1,11 +1,12 @@
 from __future__ import print_function, division, absolute_import
-from llvm.core import Type, Constant
-import llvm.core as lc
-import llvm.ee as le
-from llvm import LLVMException
+
+import llvmlite.binding as ll
+from llvmlite.llvmpy.core import Type, Constant, LLVMException
+import llvmlite.llvmpy.core as lc
+
 from numba.config import PYVERSION
 import numba.ctypes_support as ctypes
-from numba import types, utils, cgutils, _numpyadapt, _helperlib, assume
+from numba import types, utils, cgutils, _helperlib, assume
 
 _PyNone = ctypes.c_ssize_t(id(None))
 
@@ -19,23 +20,29 @@ def fix_python_api():
     """
     Execute once to install special symbols into the LLVM symbol table
     """
-    c_helpers = _helperlib.c_helpers
-    le.dylib_add_symbol("Py_None", ctypes.addressof(_PyNone))
-    le.dylib_add_symbol("NumbaArrayAdaptor", _numpyadapt.get_ndarray_adaptor())
-    le.dylib_add_symbol("NumbaComplexAdaptor",
-                        c_helpers["complex_adaptor"])
-    le.dylib_add_symbol("NumbaNativeError", id(NativeError))
-    le.dylib_add_symbol("NumbaExtractRecordData",
-                        c_helpers["extract_record_data"])
-    le.dylib_add_symbol("NumbaReleaseRecordBuffer",
-                        c_helpers["release_record_buffer"])
-    le.dylib_add_symbol("NumbaRecreateRecord",
-                        c_helpers["recreate_record"])
-    le.dylib_add_symbol("PyExc_NameError", id(NameError))
 
+    ll.add_symbol("Py_None", ctypes.addressof(_PyNone))
+    ll.add_symbol("numba_native_error", id(NativeError))
+
+    # Add C helper functions
+    c_helpers = _helperlib.c_helpers
+    for py_name in c_helpers:
+        c_name = "numba_" + py_name
+        c_address = c_helpers[py_name]
+        ll.add_symbol(c_name, c_address)
+
+    # Add all built-in exception classes
+    for obj in utils.builtins.__dict__.values():
+        if isinstance(obj, type) and issubclass(obj, BaseException):
+            ll.add_symbol("PyExc_%s" % (obj.__name__), id(obj))
 
 
 class PythonAPI(object):
+    """
+    Code generation facilities to call into the CPython C API (and related
+    helpers).
+    """
+
     def __init__(self, context, builder):
         """
         Note: Maybe called multiple times when lowering a function
@@ -54,9 +61,14 @@ class PythonAPI(object):
         self.double = Type.double()
         self.py_ssize_t = self.context.get_value_type(types.intp)
         self.cstring = Type.pointer(Type.int(8))
+        self.gil_state = Type.int(_helperlib.py_gil_state_size * 8)
 
     # ------ Python API -----
 
+    #
+    # Basic object API
+    #
+
     def incref(self, obj):
         fnty = Type.function(Type.void(), [self.pyobj])
         fn = self._get_function(fnty, name="Py_IncRef")
@@ -67,6 +79,10 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="Py_DecRef")
         self.builder.call(fn, [obj])
 
+    #
+    # Argument unpacking
+    #
+
     def parse_tuple_and_keywords(self, args, kws, fmt, keywords, *objs):
         charptr = Type.pointer(Type.int(8))
         charptrary = Type.pointer(charptr)
@@ -82,13 +98,9 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="PyArg_ParseTuple")
         return self.builder.call(fn, [args, fmt] + list(objs))
 
-    def dict_getitem_string(self, dic, name):
-        """Returns a borrowed reference
-        """
-        fnty = Type.function(self.pyobj, [self.pyobj, self.cstring])
-        fn = self._get_function(fnty, name="PyDict_GetItemString")
-        cstr = self.context.insert_const_string(self.module, name)
-        return self.builder.call(fn, [dic, cstr])
+    #
+    # Exception handling
+    #
 
     def err_occurred(self):
         fnty = Type.function(self.pyobj, ())
@@ -103,6 +115,10 @@ class PythonAPI(object):
     def err_set_string(self, exctype, msg):
         fnty = Type.function(Type.void(), [self.pyobj, self.cstring])
         fn = self._get_function(fnty, name="PyErr_SetString")
+        if isinstance(exctype, str):
+            exctype = self.get_c_object(exctype)
+        if isinstance(msg, str):
+            msg = self.context.insert_const_string(self.module, msg)
         return self.builder.call(fn, (exctype, msg))
 
     def err_set_object(self, exctype, excval):
@@ -110,22 +126,95 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="PyErr_SetObject")
         return self.builder.call(fn, (exctype, excval))
 
-    def import_module_noblock(self, modname):
-        fnty = Type.function(self.pyobj, [self.cstring])
-        fn = self._get_function(fnty, name="PyImport_ImportModuleNoBlock")
-        return self.builder.call(fn, [modname])
+    def raise_native_error(self, msg):
+        cstr = self.context.insert_const_string(self.module, msg)
+        self.err_set_string(self.native_error_type, cstr)
 
-    def call_function_objargs(self, callee, objargs):
-        fnty = Type.function(self.pyobj, [self.pyobj], var_arg=True)
-        fn = self._get_function(fnty, name="PyObject_CallFunctionObjArgs")
-        args = [callee] + list(objargs)
-        args.append(self.context.get_constant_null(types.pyobject))
-        return self.builder.call(fn, args)
+    def raise_exception(self, exctype, excval):
+        # XXX This produces non-reusable bitcode: the pointer's value
+        # is specific to this process execution.
+        exctypeaddr = self.context.get_constant(types.intp, id(exctype))
+        excvaladdr = self.context.get_constant(types.intp, id(excval))
+        self.err_set_object(exctypeaddr.inttoptr(self.pyobj),
+                            excvaladdr.inttoptr(self.pyobj))
 
-    def call(self, callee, args, kws):
-        fnty = Type.function(self.pyobj, [self.pyobj] * 3)
-        fn = self._get_function(fnty, name="PyObject_Call")
-        return self.builder.call(fn, (callee, args, kws))
+    def get_c_object(self, name):
+        """
+        Get a Python object through its C-accessible *name*.
+        (e.g. "PyExc_ValueError").
+        """
+        try:
+            gv = self.module.get_global_variable_named(name)
+        except LLVMException:
+            gv = self.module.add_global_variable(self.pyobj.pointee, name)
+        return gv
+
+    @property
+    def native_error_type(self):
+        return self.get_c_object("numba_native_error")
+
+    def raise_missing_global_error(self, name):
+        msg = "global name '%s' is not defined" % name
+        cstr = self.context.insert_const_string(self.module, msg)
+        self.err_set_string("PyExc_NameError", cstr)
+
+    def raise_missing_name_error(self, name):
+        msg = "name '%s' is not defined" % name
+        cstr = self.context.insert_const_string(self.module, msg)
+        self.err_set_string("PyExc_NameError", cstr)
+
+    #
+    # Concrete dict API
+    #
+
+    def dict_getitem_string(self, dic, name):
+        """Returns a borrowed reference
+        """
+        fnty = Type.function(self.pyobj, [self.pyobj, self.cstring])
+        fn = self._get_function(fnty, name="PyDict_GetItemString")
+        cstr = self.context.insert_const_string(self.module, name)
+        return self.builder.call(fn, [dic, cstr])
+
+    def dict_new(self, presize=0):
+        if presize == 0:
+            fnty = Type.function(self.pyobj, ())
+            fn = self._get_function(fnty, name="PyDict_New")
+            return self.builder.call(fn, ())
+        else:
+            fnty = Type.function(self.pyobj, [self.py_ssize_t])
+            fn = self._get_function(fnty, name="_PyDict_NewPresized")
+            return self.builder.call(fn,
+                                     [Constant.int(self.py_ssize_t, presize)])
+
+    def dict_setitem(self, dictobj, nameobj, valobj):
+        fnty = Type.function(Type.int(), (self.pyobj, self.pyobj,
+                                          self.pyobj))
+        fn = self._get_function(fnty, name="PyDict_SetItem")
+        return self.builder.call(fn, (dictobj, nameobj, valobj))
+
+    def dict_setitem_string(self, dictobj, name, valobj):
+        fnty = Type.function(Type.int(), (self.pyobj, self.cstring,
+                                          self.pyobj))
+        fn = self._get_function(fnty, name="PyDict_SetItemString")
+        cstr = self.context.insert_const_string(self.module, name)
+        return self.builder.call(fn, (dictobj, cstr, valobj))
+
+    def dict_pack(self, keyvalues):
+        """
+        Args
+        -----
+        keyvalues: iterable of (str, llvm.Value of PyObject*)
+        """
+        dictobj = self.dict_new()
+        not_null = cgutils.is_not_null(self.builder, dictobj)
+        with cgutils.if_likely(self.builder, not_null):
+            for k, v in keyvalues:
+                self.dict_setitem_string(dictobj, k, v)
+        return dictobj
+
+    #
+    # Concrete number APIs
+    #
 
     def float_from_double(self, fval):
         fnty = Type.function(self.pyobj, [self.double])
@@ -213,58 +302,53 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="PyNumber_%s" % name)
         return fn
 
-    def number_add(self, lhs, rhs):
-        fn = self._get_number_operator("Add")
+    def _call_number_operator(self, name, lhs, rhs, inplace=False):
+        if inplace:
+            name = "InPlace" + name
+        fn = self._get_number_operator(name)
         return self.builder.call(fn, [lhs, rhs])
 
-    def number_subtract(self, lhs, rhs):
-        fn = self._get_number_operator("Subtract")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_add(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("Add", lhs, rhs, inplace=inplace)
 
-    def number_multiply(self, lhs, rhs):
-        fn = self._get_number_operator("Multiply")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_subtract(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("Subtract", lhs, rhs, inplace=inplace)
 
-    def number_divide(self, lhs, rhs):
+    def number_multiply(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("Multiply", lhs, rhs, inplace=inplace)
+
+    def number_divide(self, lhs, rhs, inplace=False):
         assert PYVERSION < (3, 0)
-        fn = self._get_number_operator("Divide")
-        return self.builder.call(fn, [lhs, rhs])
+        return self._call_number_operator("Divide", lhs, rhs, inplace=inplace)
 
-    def number_truedivide(self, lhs, rhs):
-        fn = self._get_number_operator("TrueDivide")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_truedivide(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("TrueDivide", lhs, rhs, inplace=inplace)
 
-    def number_floordivide(self, lhs, rhs):
-        fn = self._get_number_operator("FloorDivide")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_floordivide(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("FloorDivide", lhs, rhs, inplace=inplace)
 
-    def number_remainder(self, lhs, rhs):
-        fn = self._get_number_operator("Remainder")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_remainder(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("Remainder", lhs, rhs, inplace=inplace)
 
-    def number_lshift(self, lhs, rhs):
-        fn = self._get_number_operator("Lshift")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_lshift(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("Lshift", lhs, rhs, inplace=inplace)
 
-    def number_rshift(self, lhs, rhs):
-        fn = self._get_number_operator("Rshift")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_rshift(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("Rshift", lhs, rhs, inplace=inplace)
 
-    def number_and(self, lhs, rhs):
-        fn = self._get_number_operator("And")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_and(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("And", lhs, rhs, inplace=inplace)
 
-    def number_or(self, lhs, rhs):
-        fn = self._get_number_operator("Or")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_or(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("Or", lhs, rhs, inplace=inplace)
 
-    def number_xor(self, lhs, rhs):
-        fn = self._get_number_operator("Xor")
-        return self.builder.call(fn, [lhs, rhs])
+    def number_xor(self, lhs, rhs, inplace=False):
+        return self._call_number_operator("Xor", lhs, rhs, inplace=inplace)
 
-    def number_power(self, lhs, rhs):
+    def number_power(self, lhs, rhs, inplace=False):
         fnty = Type.function(self.pyobj, [self.pyobj] * 3)
-        fn = self._get_function(fnty, "PyNumber_Power")
+        fname = "PyNumber_InPlacePower" if inplace else "PyNumber_Power"
+        fn = self._get_function(fnty, fname)
         return self.builder.call(fn, [lhs, rhs, self.borrow_none()])
 
     def number_negative(self, obj):
@@ -292,29 +376,6 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="PyFloat_AsDouble")
         return self.builder.call(fn, [fobj])
 
-    def object_istrue(self, obj):
-        fnty = Type.function(Type.int(), [self.pyobj])
-        fn = self._get_function(fnty, name="PyObject_IsTrue")
-        return self.builder.call(fn, [obj])
-
-    def object_not(self, obj):
-        fnty = Type.function(Type.int(), [self.pyobj])
-        fn = self._get_function(fnty, name="PyObject_Not")
-        return self.builder.call(fn, [obj])
-
-    def object_richcompare(self, lhs, rhs, opstr):
-        """
-        Refer to Python source Include/object.h for macros definition
-        of the opid.
-        """
-        ops = ['<', '<=', '==', '!=', '>', '>=']
-        opid = ops.index(opstr)
-        assert 0 <= opid < len(ops)
-        fnty = Type.function(self.pyobj, [self.pyobj, self.pyobj, Type.int()])
-        fn = self._get_function(fnty, name="PyObject_RichCompare")
-        lopid = self.context.get_constant(types.int32, opid)
-        return self.builder.call(fn, (lhs, rhs, lopid))
-
     def bool_from_bool(self, bval):
         """
         Get a Python bool from a LLVM boolean.
@@ -342,6 +403,181 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="PyComplex_ImagAsDouble")
         return self.builder.call(fn, [cobj])
 
+    #
+    # List and sequence APIs
+    #
+
+    def sequence_getslice(self, obj, start, stop):
+        fnty = Type.function(self.pyobj, [self.pyobj, self.py_ssize_t,
+                                          self.py_ssize_t])
+        fn = self._get_function(fnty, name="PySequence_GetSlice")
+        return self.builder.call(fn, (obj, start, stop))
+
+    def sequence_tuple(self, obj):
+        fnty = Type.function(self.pyobj, [self.pyobj])
+        fn = self._get_function(fnty, name="PySequence_Tuple")
+        return self.builder.call(fn, [obj])
+
+    def list_new(self, szval):
+        fnty = Type.function(self.pyobj, [self.py_ssize_t])
+        fn = self._get_function(fnty, name="PyList_New")
+        return self.builder.call(fn, [szval])
+
+    def list_setitem(self, seq, idx, val):
+        """
+        Warning: Steals reference to ``val``
+        """
+        fnty = Type.function(Type.int(), [self.pyobj, self.py_ssize_t,
+                                          self.pyobj])
+        fn = self._get_function(fnty, name="PyList_SetItem")
+        return self.builder.call(fn, [seq, idx, val])
+
+    def list_getitem(self, lst, idx):
+        """
+        Returns a borrowed reference.
+        """
+        fnty = Type.function(self.pyobj, [self.pyobj, self.py_ssize_t])
+        fn = self._get_function(fnty, name="PyList_GetItem")
+        if isinstance(idx, int):
+            idx = self.context.get_constant(types.intp, idx)
+        return self.builder.call(fn, [lst, idx])
+
+    #
+    # Concrete tuple API
+    #
+
+    def tuple_getitem(self, tup, idx):
+        """
+        Borrow reference
+        """
+        fnty = Type.function(self.pyobj, [self.pyobj, self.py_ssize_t])
+        fn = self._get_function(fnty, name="PyTuple_GetItem")
+        idx = self.context.get_constant(types.intp, idx)
+        return self.builder.call(fn, [tup, idx])
+
+    def tuple_pack(self, items):
+        fnty = Type.function(self.pyobj, [self.py_ssize_t], var_arg=True)
+        fn = self._get_function(fnty, name="PyTuple_Pack")
+        n = self.context.get_constant(types.intp, len(items))
+        args = [n]
+        args.extend(items)
+        return self.builder.call(fn, args)
+
+    def tuple_size(self, tup):
+        fnty = Type.function(self.py_ssize_t, [self.pyobj])
+        fn = self._get_function(fnty, name="PyTuple_Size")
+        return self.builder.call(fn, [tup])
+
+    def tuple_new(self, count):
+        fnty = Type.function(self.pyobj, [Type.int()])
+        fn = self._get_function(fnty, name='PyTuple_New')
+        return self.builder.call(fn, [self.context.get_constant(types.int32,
+                                                                count)])
+
+    def tuple_setitem(self, tuple_val, index, item):
+        """
+        Steals a reference to `item`.
+        """
+        fnty = Type.function(Type.int(), [self.pyobj, Type.int(), self.pyobj])
+        setitem_fn = self._get_function(fnty, name='PyTuple_SetItem')
+        index = self.context.get_constant(types.int32, index)
+        self.builder.call(setitem_fn, [tuple_val, index, item])
+
+    #
+    # Concrete set API
+    #
+
+    def set_new(self, iterable=None):
+        if iterable is None:
+            iterable = self.get_null_object()
+        fnty = Type.function(self.pyobj, [self.pyobj])
+        fn = self._get_function(fnty, name="PySet_New")
+        return self.builder.call(fn, [iterable])
+
+    def set_add(self, set, value):
+        fnty = Type.function(Type.int(), [self.pyobj, self.pyobj])
+        fn = self._get_function(fnty, name="PySet_Add")
+        return self.builder.call(fn, [set, value])
+
+    #
+    # GIL APIs
+    #
+
+    def gil_ensure(self):
+        """
+        Ensure the GIL is acquired.
+        The returned value must be consumed by gil_release().
+        """
+        gilptrty = Type.pointer(self.gil_state)
+        fnty = Type.function(Type.void(), [gilptrty])
+        fn = self._get_function(fnty, "numba_gil_ensure")
+        gilptr = cgutils.alloca_once(self.builder, self.gil_state)
+        self.builder.call(fn, [gilptr])
+        return gilptr
+
+    def gil_release(self, gil):
+        """
+        Release the acquired GIL by gil_ensure().
+        Must be pair with a gil_ensure().
+        """
+        gilptrty = Type.pointer(self.gil_state)
+        fnty = Type.function(Type.void(), [gilptrty])
+        fn = self._get_function(fnty, "numba_gil_release")
+        return self.builder.call(fn, [gil])
+
+    #
+    # Other APIs (organize them better!)
+    #
+
+    def import_module_noblock(self, modname):
+        fnty = Type.function(self.pyobj, [self.cstring])
+        fn = self._get_function(fnty, name="PyImport_ImportModuleNoBlock")
+        return self.builder.call(fn, [modname])
+
+    def call_function_objargs(self, callee, objargs):
+        fnty = Type.function(self.pyobj, [self.pyobj], var_arg=True)
+        fn = self._get_function(fnty, name="PyObject_CallFunctionObjArgs")
+        args = [callee] + list(objargs)
+        args.append(self.context.get_constant_null(types.pyobject))
+        return self.builder.call(fn, args)
+
+    def call(self, callee, args, kws):
+        fnty = Type.function(self.pyobj, [self.pyobj] * 3)
+        fn = self._get_function(fnty, name="PyObject_Call")
+        return self.builder.call(fn, (callee, args, kws))
+
+    def object_istrue(self, obj):
+        fnty = Type.function(Type.int(), [self.pyobj])
+        fn = self._get_function(fnty, name="PyObject_IsTrue")
+        return self.builder.call(fn, [obj])
+
+    def object_not(self, obj):
+        fnty = Type.function(Type.int(), [self.pyobj])
+        fn = self._get_function(fnty, name="PyObject_Not")
+        return self.builder.call(fn, [obj])
+
+    def object_richcompare(self, lhs, rhs, opstr):
+        """
+        Refer to Python source Include/object.h for macros definition
+        of the opid.
+        """
+        ops = ['<', '<=', '==', '!=', '>', '>=']
+        if opstr in ops:
+            opid = ops.index(opstr)
+            fnty = Type.function(self.pyobj, [self.pyobj, self.pyobj, Type.int()])
+            fn = self._get_function(fnty, name="PyObject_RichCompare")
+            lopid = self.context.get_constant(types.int32, opid)
+            return self.builder.call(fn, (lhs, rhs, lopid))
+        elif opstr == 'is':
+            bitflag = self.builder.icmp(lc.ICMP_EQ, lhs, rhs)
+            return self.from_native_value(bitflag, types.boolean)
+        elif opstr == 'is not':
+            bitflag = self.builder.icmp(lc.ICMP_NE, lhs, rhs)
+            return self.from_native_value(bitflag, types.boolean)
+        else:
+            raise NotImplementedError("Unknown operator {op!r}".format(
+                op=opstr))
+
     def iter_next(self, iterobj):
         fnty = Type.function(self.pyobj, [self.pyobj])
         fn = self._get_function(fnty, name="PyIter_Next")
@@ -358,6 +594,17 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="PyObject_GetAttrString")
         return self.builder.call(fn, [obj, cstr])
 
+    def object_setattr_string(self, obj, attr, val):
+        cstr = self.context.insert_const_string(self.module, attr)
+        fnty = Type.function(Type.int(), [self.pyobj, self.cstring, self.pyobj])
+        fn = self._get_function(fnty, name="PyObject_SetAttrString")
+        return self.builder.call(fn, [obj, cstr, val])
+
+    def object_delattr_string(self, obj, attr):
+        # PyObject_DelAttrString() is actually a C macro calling
+        # PyObject_SetAttrString() with value == NULL.
+        return self.object_setattr_string(obj, attr, self.get_null_object())
+
     def object_getitem(self, obj, key):
         fnty = Type.function(self.pyobj, [self.pyobj, self.pyobj])
         fn = self._get_function(fnty, name="PyObject_GetItem")
@@ -368,17 +615,6 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="PyObject_SetItem")
         return self.builder.call(fn, (obj, key, val))
 
-    def sequence_getslice(self, obj, start, stop):
-        fnty = Type.function(self.pyobj, [self.pyobj, self.py_ssize_t,
-                                          self.py_ssize_t])
-        fn = self._get_function(fnty, name="PySequence_GetSlice")
-        return self.builder.call(fn, (obj, start, stop))
-
-    def sequence_tuple(self, obj):
-        fnty = Type.function(self.pyobj, [self.pyobj])
-        fn = self._get_function(fnty, name="PySequence_Tuple")
-        return self.builder.call(fn, [obj])
-
     def string_as_string(self, strobj):
         fnty = Type.function(self.cstring, [self.pyobj])
         if PYVERSION >= (3, 0):
@@ -411,78 +647,6 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="PyObject_Str")
         return self.builder.call(fn, [obj])
 
-    def tuple_getitem(self, tup, idx):
-        """
-        Borrow reference
-        """
-        fnty = Type.function(self.pyobj, [self.pyobj, self.py_ssize_t])
-        fn = self._get_function(fnty, name="PyTuple_GetItem")
-        idx = self.context.get_constant(types.intp, idx)
-        return self.builder.call(fn, [tup, idx])
-
-    def tuple_pack(self, items):
-        fnty = Type.function(self.pyobj, [self.py_ssize_t], var_arg=True)
-        fn = self._get_function(fnty, name="PyTuple_Pack")
-        n = self.context.get_constant(types.intp, len(items))
-        args = [n]
-        args.extend(items)
-        return self.builder.call(fn, args)
-
-    def tuple_size(self, tup):
-        fnty = Type.function(self.py_ssize_t, [self.pyobj])
-        fn = self._get_function(fnty, name="PyTuple_Size")
-        return self.builder.call(fn, [tup])
-
-    def list_new(self, szval):
-        fnty = Type.function(self.pyobj, [self.py_ssize_t])
-        fn = self._get_function(fnty, name="PyList_New")
-        return self.builder.call(fn, [szval])
-
-    def list_setitem(self, seq, idx, val):
-        """
-        Warning: Steals reference to ``val``
-        """
-        fnty = Type.function(Type.int(), [self.pyobj, self.py_ssize_t,
-                                          self.pyobj])
-        fn = self._get_function(fnty, name="PyList_SetItem")
-        return self.builder.call(fn, [seq, idx, val])
-
-    def set_new(self, iterable=None):
-        if iterable is None:
-            iterable = self.get_null_object()
-        fnty = Type.function(self.pyobj, [self.pyobj])
-        fn = self._get_function(fnty, name="PySet_New")
-        return self.builder.call(fn, [iterable])
-
-    def set_add(self, set, value):
-        fnty = Type.function(Type.int(), [self.pyobj, self.pyobj])
-        fn = self._get_function(fnty, name="PySet_Add")
-        return self.builder.call(fn, [set, value])
-
-    def dict_new(self, presize=0):
-        if presize == 0:
-            fnty = Type.function(self.pyobj, ())
-            fn = self._get_function(fnty, name="PyDict_New")
-            return self.builder.call(fn, ())
-        else:
-            fnty = Type.function(self.pyobj, [self.py_ssize_t])
-            fn = self._get_function(fnty, name="_PyDict_NewPresized")
-            return self.builder.call(fn,
-                                     [Constant.int(self.py_ssize_t, presize)])
-
-    def dict_setitem(self, dictobj, nameobj, valobj):
-        fnty = Type.function(Type.int(), (self.pyobj, self.pyobj,
-                                          self.pyobj))
-        fn = self._get_function(fnty, name="PyDict_SetItem")
-        return self.builder.call(fn, (dictobj, nameobj, valobj))
-
-    def dict_setitem_string(self, dictobj, name, valobj):
-        fnty = Type.function(Type.int(), (self.pyobj, self.cstring,
-                                          self.pyobj))
-        fn = self._get_function(fnty, name="PyDict_SetItemString")
-        cstr = self.context.insert_const_string(self.module, name)
-        return self.builder.call(fn, (dictobj, cstr, valobj))
-
     def make_none(self):
         obj = self._get_object("Py_None")
         self.incref(obj)
@@ -497,6 +661,14 @@ class PythonAPI(object):
         fn = self._get_function(fnty, name="PySys_WriteStdout")
         return self.builder.call(fn, (fmt,) + args)
 
+    def object_dump(self, obj):
+        """
+        Dump a Python object on C stderr.  For debugging purposes.
+        """
+        fnty = Type.function(Type.void(), [self.pyobj])
+        fn = self._get_function(fnty, name="_PyObject_Dump")
+        return self.builder.call(fn, (obj,))
+
     # ------ utils -----
 
     def _get_object(self, name):
@@ -541,19 +713,6 @@ class PythonAPI(object):
                 self.list_setitem(seq, idx, items[i])
         return seq
 
-    def dict_pack(self, keyvalues):
-        """
-        Args
-        -----
-        keyvalues: iterable of (str, llvm.Value of PyObject*)
-        """
-        dictobj = self.dict_new()
-        not_null = cgutils.is_not_null(self.builder, dictobj)
-        with cgutils.if_likely(self.builder, not_null):
-            for k, v in keyvalues:
-                self.dict_setitem_string(dictobj, k, v)
-        return dictobj
-
     def to_native_arg(self, obj, typ):
         if isinstance(typ, types.Record):
             # Generate a dummy integer type that has the size of Py_buffer
@@ -646,9 +805,32 @@ class PythonAPI(object):
             else:
                 return cplx._getvalue()
 
+        elif isinstance(typ, types.NPDatetime):
+            val = self.extract_np_datetime(obj)
+            return val
+
+        elif isinstance(typ, types.NPTimedelta):
+            val = self.extract_np_timedelta(obj)
+            return val
+
         elif isinstance(typ, types.Array):
             return self.to_native_array(typ, obj)
 
+        elif isinstance(typ, types.Optional):
+            isnone = self.builder.icmp(lc.ICMP_EQ, obj, self.borrow_none())
+            with cgutils.ifelse(self.builder, isnone) as (then, orelse):
+                with then:
+                    noneval = self.context.make_optional_none(self.builder, typ.type)
+                    ret = cgutils.alloca_once(self.builder, noneval.type)
+                    self.builder.store(noneval, ret)
+
+                with orelse:
+                    val = self.to_native_value(obj, typ.type)
+                    just = self.context.make_optional_value(self.builder,
+                                                            typ.type, val)
+                    self.builder.store(just, ret)
+            return ret
+
         raise NotImplementedError(typ)
 
     def from_native_return(self, val, typ):
@@ -691,6 +873,12 @@ class PythonAPI(object):
                                       types.float32, types.float64)
             return self.complex_from_doubles(freal, fimag)
 
+        elif isinstance(typ, types.NPDatetime):
+            return self.create_np_datetime(val, typ.unit_code)
+
+        elif isinstance(typ, types.NPTimedelta):
+            return self.create_np_timedelta(val, typ.unit_code)
+
         elif typ == types.none:
             ret = self.make_none()
             return ret
@@ -713,8 +901,8 @@ class PythonAPI(object):
             dtypeobj = dtype_addr.inttoptr(self.pyobj)
             return self.recreate_record(ptr, size, dtypeobj)
 
-        elif isinstance(typ, types.UniTuple):
-            return self.from_unituple(typ, val)
+        elif isinstance(typ, (types.Tuple, types.UniTuple)):
+            return self.from_tuple(typ, val)
 
         raise NotImplementedError(typ)
 
@@ -742,109 +930,68 @@ class PythonAPI(object):
         self.incref(parent)
         return parent
 
-    def from_unituple(self, typ, val):
+    def from_tuple(self, typ, val):
         tuple_val = self.tuple_new(typ.count)
 
-        for i in range(typ.count):
+        for i, dtype in enumerate(typ):
             item = self.builder.extract_value(val, i)
-            obj = self.from_native_value(item, typ.dtype)
+            obj = self.from_native_value(item,  dtype)
             self.tuple_setitem(tuple_val, i, obj)
 
         return tuple_val
 
-    def tuple_new(self, count):
-        fnty = Type.function(self.pyobj, [Type.int()])
-        fn = self._get_function(fnty, name='PyTuple_New')
-        return self.builder.call(fn, [self.context.get_constant(types.int32,
-                                                                count)])
-
-    def tuple_setitem(self, tuple_val, index, item):
-        """
-        Steals a reference to `item`.
-        """
-        fnty = Type.function(Type.int(), [self.pyobj, Type.int(), self.pyobj])
-        setitem_fn = self._get_function(fnty, name='PyTuple_SetItem')
-        index = self.context.get_constant(types.int32, index)
-        self.builder.call(setitem_fn, [tuple_val, index, item])
-
     def numba_array_adaptor(self, ary, ptr):
         voidptr = Type.pointer(Type.int(8))
         fnty = Type.function(Type.int(), [self.pyobj, voidptr])
-        fn = self._get_function(fnty, name="NumbaArrayAdaptor")
+        fn = self._get_function(fnty, name="numba_adapt_ndarray")
         fn.args[0].add_attribute(lc.ATTR_NO_CAPTURE)
         fn.args[1].add_attribute(lc.ATTR_NO_CAPTURE)
         return self.builder.call(fn, (ary, ptr))
 
     def complex_adaptor(self, cobj, cmplx):
         fnty = Type.function(Type.int(), [self.pyobj, cmplx.type])
-        fn = self._get_function(fnty, name="NumbaComplexAdaptor")
+        fn = self._get_function(fnty, name="numba_complex_adaptor")
         return self.builder.call(fn, [cobj, cmplx])
 
     def extract_record_data(self, obj, pbuf):
         fnty = Type.function(self.voidptr, [self.pyobj,
                                                          self.voidptr])
-        fn = self._get_function(fnty, name="NumbaExtractRecordData")
+        fn = self._get_function(fnty, name="numba_extract_record_data")
         return self.builder.call(fn, [obj, pbuf])
 
     def release_record_buffer(self, pbuf):
         fnty = Type.function(Type.void(), [self.voidptr])
-        fn = self._get_function(fnty, name="NumbaReleaseRecordBuffer")
+        fn = self._get_function(fnty, name="numba_release_record_buffer")
         return self.builder.call(fn, [pbuf])
 
+    def extract_np_datetime(self, obj):
+        fnty = Type.function(Type.int(64), [self.pyobj])
+        fn = self._get_function(fnty, name="numba_extract_np_datetime")
+        return self.builder.call(fn, [obj])
+
+    def extract_np_timedelta(self, obj):
+        fnty = Type.function(Type.int(64), [self.pyobj])
+        fn = self._get_function(fnty, name="numba_extract_np_timedelta")
+        return self.builder.call(fn, [obj])
+
+    def create_np_datetime(self, val, unit_code):
+        unit_code = Constant.int(Type.int(), unit_code)
+        fnty = Type.function(self.pyobj, [Type.int(64), Type.int()])
+        fn = self._get_function(fnty, name="numba_create_np_datetime")
+        return self.builder.call(fn, [val, unit_code])
+
+    def create_np_timedelta(self, val, unit_code):
+        unit_code = Constant.int(Type.int(), unit_code)
+        fnty = Type.function(self.pyobj, [Type.int(64), Type.int()])
+        fn = self._get_function(fnty, name="numba_create_np_timedelta")
+        return self.builder.call(fn, [val, unit_code])
+
     def recreate_record(self, pdata, size, dtypeaddr):
         fnty = Type.function(self.pyobj, [Type.pointer(Type.int(8)),
                                           Type.int(), self.pyobj])
-        fn = self._get_function(fnty, name="NumbaRecreateRecord")
+        fn = self._get_function(fnty, name="numba_recreate_record")
         return self.builder.call(fn, [pdata, size, dtypeaddr])
 
-    def get_module_dict_symbol(self):
-        md_pymod = cgutils.MetadataKeyStore(self.module, "python.module")
-        pymodname = ".pymodule.dict." + md_pymod.get()
-
-        try:
-            gv = self.module.get_global_variable_named(name=pymodname)
-        except LLVMException:
-            gv = self.module.add_global_variable(self.pyobj.pointee,
-                                                 name=pymodname)
-        return gv
-
-    def get_module_dict(self):
-        return self.get_module_dict_symbol()
-        # return self.builder.load(gv)
-
-    def raise_native_error(self, msg):
-        cstr = self.context.insert_const_string(self.module, msg)
-        self.err_set_string(self.native_error_type, cstr)
-
-    def raise_exception(self, exctype, excval):
-        exctypeaddr = self.context.get_constant(types.intp, id(exctype))
-        excvaladdr = self.context.get_constant(types.intp, id(excval))
-        self.err_set_object(exctypeaddr.inttoptr(self.pyobj),
-                            excvaladdr.inttoptr(self.pyobj))
-
-    @property
-    def native_error_type(self):
-        name = "NumbaNativeError"
-        try:
-            return self.module.get_global_variable_named(name)
-        except LLVMException:
-            return self.module.add_global_variable(self.pyobj.pointee,
-                                                   name=name)
-
-    def raise_missing_global_error(self, name):
-        msg = "global name '%s' is not defined" % name
-        cstr = self.context.insert_const_string(self.module, msg)
-        self.err_set_string(self.name_error_type, cstr)
-
-    @property
-    def name_error_type(self):
-        name = "PyExc_NameError"
-        try:
-            return self.module.get_global_variable_named(name)
-        except LLVMException:
-            return self.module.add_global_variable(self.pyobj.pointee,
-                                                   name=name)
-
     def string_from_constant_string(self, string):
         cstr = self.context.insert_const_string(self.module, string)
         sz = self.context.get_constant(types.intp, len(string))
